"""This is a quickly hacked module for the Sopel IRC bot. It executes the dict
command line client to query available dictionaries. Using a proper dict library
would be far better and has not been done due to time constraints.

(C) 2017 Sebastian Humenda, licensed under the terms of the GPL, version 3, or
any later version, published by the FSF."""


import os
import random
import re
import subprocess
import sys

from sopel import module

dict_pattern = re.compile(r"^[a-z]{3}-[a-z]{3}$")
dict_file_pattern = re.compile(r'^freedict-([a-z]{3}-[a-z]{3}).index$')


def answer(bot, trigger, msg):
    if trigger.is_privmsg:
        bot.say(msg)
    else:
        bot.reply(msg)

def error(bot, msg):
    """Print an error message"""
    bot.reply(msg)


def lookup(bot, language, word):
    if language:
        if not dict_pattern.search(language):
            error(bot, "Invalid dictionary specifier.")
            return
    if len(word) > 20:
        error(bot, "Sorry, this word is too long.")
        return
    proc = None
    if language:
        proc = subprocess.Popen(['dict',  '-d', 'fd-' + language,
            word.encode(sys.getdefaultencoding())], stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    else:
        proc = subprocess.Popen(['dict', word.encode(sys.getdefaultencoding())],
                stdout=subprocess.PIPE, stderr=subprocess.PIPE)

    output = proc.communicate()
    if proc.wait():
        # try to get alternate suggestions
        stderr = output[1].decode(sys.getdefaultencoding())
        if 'perhaps you mean' in stderr:
            suggestions = '; '.join(stderr.split('\n')[1:])
            suggestions = suggestions[suggestions.find(':')+2:]
            bot.say("Sorry, entry not found, try: " + suggestions)
        else:
            bot.say("Sorry, but unfortunately this entry is not available.")
        return
    else:
        # strip first two lines
        text = output[0].decode(sys.getdefaultencoding()).lstrip()
        text = (l for l in text.split('\n')[3:] if l)
        text = [line for line in text if not line.lstrip().startswith("From ")]
        if len(text) > 5: # ToDo: use paste service
            text = text[:6]
            text[-1] += '  … a few more lines omitted.'
            if not language:
                text[-1] = text[-1][:-1] + ', try restricting the search by ' +\
                    'specifying a dictionary.'
        for line in text:
            bot.say(line)

################################################################################
# actual triggers

@module.rate(server=0.4)
@module.commands('dict')
def dict_lookup(bot, trigger):
    def report_invalid_input():
        answer(bot, trigger, ("Usage: .dict [DICTIONARY] word…; try .help dict "
            "for an explanation."))

    if not trigger.group(2):
        report_invalid_input()
        return

    data = trigger.group(2).split(' ')
    word = language = None
    if len(data) == 0:
        report_invalid_input()
        return
    elif len(data) == 1:
        word = data[0]
    else:
        language = data[0]
        word = ' '.join(data[1:])

    lookup(bot, language, word)


@module.rate(server=0.2)
@module.commands('list')
def list_dicts(bot, trigger):
    files = ', '.join(sorted(map(lambda d: d.groups()[0],
            filter(bool, map(dict_file_pattern.search,
                os.listdir('/usr/share/dictd'))))))
    answer(bot, trigger, "Available databases: " + files)


@module.commands('help')
def help_user(bot, trigger):
    args = trigger.group(2)
    if args:
        if args == '.list' or args == 'list':
            answer(bot, trigger, "Type .list and I will present you with a list of all FreeDict dictionaries.")
        elif args == '.dict' or args == 'dict':
            answer(bot, trigger, ("Type .dict <dictionary> <phrase> to "
                "look up something in the desired dictionary (see .list)."))
            answer(bot, trigger, "Alternatively, you can try .dict <word>, " + \
                    "but I'll only output the first matches, if too many dictionaries contain this word.")
            answer(bot, trigger, "For instance, why not try: " + \
                    random.choice(['.dict lat-deu sapientia',
                    '.dict deu-eng Hase',
                    '.dict fra-eng ordinateur',
                    '.dict eng-spa house']))
        else:
            answer(bot, trigger, "I am afraid this is something that " + \
                    "I am not entitled to know.")
    else: # general help
        answer(bot, trigger, ("I'm serving your dictionary queries. Try "
            "the command .list to get a list of dictionaries and the command "
            ".dict to look up a word. Use .help <command> to get information on the usage."))





"""JSON output writer."""

import json

def write_freedict_database(path, dicts, tools):
    """Write a freedict database to ''path`` in JSON format."""
    serialized = []
    for dictionary in dicts:
        if not dictionary.is_complete():
            raise ValueError("Not all mandatory keys set.")
        essence = {k: v  for k, v in dictionary.get_attributes().items()
                if not v is None}
        essence['name'] = dictionary.get_name()
        # set maintainer if none present
        if 'maintainerName' not in dictionary or not dictionary['maintainerName']:
            essence['maintainerName'] = 'FreeDict - no maintainer assigned'
        essence['releases'] = []
        for release in dictionary.get_downloads():
            essence['releases'].append({'platform': str(release.format),
                'size': str(release.size),
                'date': release.last_modification_date,
                'URL': str(release), 'version': str(release.version),
                'checksum': release.hash,
            })
        serialized.append(essence)
    serialized.append({'software': {'tools': tools}})
    with open(path, 'w', encoding='utf-8') as fhandle:
        json.dump(serialized, fhandle, indent=2, sort_keys=True)




"""Constants for the freedict-API generator."""

import re

# domain where downloads are stored
PROJECTHOME_HOST = 'download.freedict.org'
# base on web server, used for the link generation
RELEASE_HTTP_BASE = '/dictionaries'
# base for the freedict tools releases, **ful URL**
RELEASE_HTTP_TOOL_BASE = 'https://download.freedict.org/tools/'

# pattern to identify dictionaries; matches three-digit ISO 6639 letter codes
DICTIONARY_PATTERN = re.compile(r'(?:freedict-)?([a-z]{3}-[a-z]{3}).*')

# Pattern to identify directory names of directories; matches English names of
# languages, separated by " - "
DICTIONARY_DIRECTORY_PATTERN = re.compile(r"[A-Z][a-z]+%20-%20[A-z][a-z]+")

# Identification of versions; basically distutils.version.StrictVersion, but
# permits "-" as separator, too




#!/usr/bin/env python3
#vim: set expandtab sts=4 ts=4 sw=4 autoindent ft=python:
"""This script downloads all (already approved) dictionary from the WikDict
project to be included in FreeDict's repository.
This script requires a FreeDict API file to figure out which dictionaries are
from WikDict and which not. It searches first for a local file referenced by the
FreeDict configuration. This is the preferred way. Then it downloads a version
from freedict.org.
"""

#pylint: disable=multiple-imports
import enum
import html.parser
import json
import multiprocessing
import os
import re
import sys
import urllib.request, urllib.parse
import shutil
import xml.etree.ElementTree as ET
from datetime import date

SOURCE_URL = 'https://download.wikdict.com/dictionaries/tei/no-infl/'
# minimal number of words to consider a dictionary for inclusion
MIN_WORD_COUNT = 10000
DOWNLOAD_PREFIX = 'http://{0.netloc}{0.path}'.format(urllib.parse.urlsplit(SOURCE_URL))
def get_fd_api():
    """Read local FreeDict API file or load from given path from configuration
    or download it from freedict.org as fallback.
    It is advised to generate a fresh API copy locally.
    This function will try to load a virtual environment, if configured."""
    js = None
    #pylint: disable=bare-except
    api_file = lambda c: os.path.join(os.path.expanduser(
            c['DEFAULT']['api_output_path']), 'freedict-database.json')

    try:
        import fd_tool.config as config
        cnf = config.discover_and_load()
        js = json.load(open(api_file(cnf), encoding="UTF-8"))
    except ImportError:
        # try to activate virtual env, if configured but not sourced yet
        paths = [os.path.join(os.path.expanduser("~"), '.config/freedict/freedictrc')]
        if os.environ.get('LOCALAPPDATA'):
            paths.append(os.path.join(os.environ['LOCALAPPDATA'], 'freedict/freedict.ini'))
            conffile = [path for path in paths if os.path.exists(path)]
            if conffile:
                import configparser
                cnf = configparser.ConfigParser()
                cnf.read_file(open(conffile[0], encoding="UTF-8"))
                if 'DEFAULT' in cnf and 'api_output_path' in cnf['DEFAULT']:
                    with open(api_file(cnf), encoding="UTF-8") as fhandle:
                        js = json.load(fhandle)
    except KeyboardInterrupt:
        pass
    if not js:
        from urllib.request import urlopen
        with urlopen('https://freedict.org/freedict-database.json') as u:
            js = json.loads(u.read().decode('utf-8'))
    # strip all non-dictionary nodes (e.g. 'software'
    return [node for node in js if 'name' in node]


class LinkExtractor(html.parser.HTMLParser):
    """Extract all URL's from a HTML file. Use extract_links for a more
    easy-to-use function."""
    def __init__(self):
        self.links = []
        super().__init__()

    def handle_starttag(self, tag, attrs):
        if tag == 'a':
            attrs = dict(attrs)
            if 'href' in attrs:
                self.links.append(attrs['href'])


def extract_links(from_string):
    """Return a list with all links contained in the HTML page passed as input
    parameter."""
    parser = LinkExtractor()
    parser.feed(from_string)
    return parser.links


def download(link):
    """Download given link and decode the bytes."""
    try:
        with urllib.request.urlopen(link) as u:
            return u.read().decode('UTF-8')
    except urllib.error.HTTPError as h:
        if int(h.code) == 404:
            reason = '%s; url: %s' % (str(h), link)
            raise urllib.error.URLError(reason)
        else:
            raise h from None


def assert_correct_working_directory():
    """Check that this script is executed from the correct directory."""
    num_files = sum(1 for fn in os.listdir('.')
            if re.search('^[a-z]{3}-[a-z]{3}$', fn))
    if num_files < 2: # less than two dictionaries, probably not a dictionary root
        print("Error: must be run from a FreeDict source root, i.e. "
                "where all generated dictionaries are stored.")
        sys.exit(9)


def make_changelog(path):
    tmpl_vars = {
            'date': date.today(),
            'dict': path,
    }
    changelog_path = 'ChangeLog'
    changelog = ''
    if os.path.exists(changelog_path):
        with open(os.path.join(path, changelog_path), 'r', encoding="UTF-8") as chlog_file:
            changelog += chlog_file.read()
    with open(os.path.join(path, 'ChangeLog'), 'w', encoding="UTF-8") as f:
        f.write(("""\n
{date}

  * automatic import of {dict} dictionary from WikDict
        \n""" + changelog).strip().format(**tmpl_vars) + '\n')


def update_dict_files(path, shared_file_path):
    dir_template = os.path.join(
            os.path.dirname(os.path.realpath(__file__)),
            'template'
    )
    shutil.rmtree(path, ignore_errors=True)
    os.makedirs(path, exist_ok=True)
    def copy(files):
        for file in files:
            try:
                shutil.copy2(file, path)
            except shutil.SameFileError:
                pass
    copy(os.path.join(shared_file_path, f) for f in
            ('freedict-dictionary.css', 'freedict-P5.dtd', 'INSTALL',
                'freedict-P5.rng', 'freedict-P5.xml'))
    copy(os.path.join(dir_template, f) for f in os.listdir(dir_template))


def dict_exists_from_other_source(api_dump, dictname):
    """FreeDict dictionaries come from different sources. At the time of writing,
these are the VCS-tracked, partly hand-written dictionaries and the
auto-imported dictionaries. WikDict are auto-imported and hence in the bucked of
non-VCS tracked, generated dictionaries. Before importing a WikDict dictionary,
we have to check whether a dictionary with the same name exists somewhere
else."""
    dictionary = [d for d in api_dump if d['name'] == dictname]

    return dictionary and ('sourceURL' not in dictionary[0] or \
            'wikdict' not in dictionary[0]['sourceURL'])

def enough_headwords(tei):
    """This function parses the TEI header and returns true if the given
    dictionary contains more than the minimal required number of headwords."""
    tei = ET.fromstring(tei)
    node = tei.find('*//{http://www.tei-c.org/ns/1.0}extent')
    count = re.search(r'(\d+\s*,?\.?\d*)\s+.*word', node.text).groups()[0]
    return int(count.strip(' ,.')) >= MIN_WORD_COUNT

def parse_links():
    with urllib.request.urlopen(SOURCE_URL) as src:
        try:
            data = src.read().decode('utf-8')
        except UnicodeEncodeError:
            raise ValueError("Could not encode page with encoding UTF-8; please adjust manually")
        return (l  for l in extract_links(data)  if l.endswith('.tei'))

class DictionaryStrategy(enum.Enum):
    TooSmall = 0
    ManuallyEdited = 1
    Imported = 2
    Rubbish = 3

def import_dictionary(api, link):
    """Import a dictionary from WikDict and prepare dictionary directory for a
    release."""
    if not urllib.parse.urlsplit(link)[1]: # no host in URL
        link = urllib.parse.urljoin(DOWNLOAD_PREFIX, link)
    base_name = os.path.splitext(link.split('/')[-1])[0] # name without .tei
    if not re.match(r'\w{3}-\w{3}', base_name):
        return (DictionaryStrategy.Rubbish, None)
    if dict_exists_from_other_source(api, base_name):
        return (base_name, DictionaryStrategy.ManuallyEdited)
    tei = download(link)
    if not enough_headwords(tei):
        return (base_name, DictionaryStrategy.TooSmall)

    print('Importing', base_name)
    if not os.path.exists(base_name):
        os.makedirs(base_name)
    # erase old files, write new ones
    update_dict_files(base_name, sys.argv[1])
    with open(os.path.join(base_name, base_name + '.tei'), 'w',
            encoding='utf-8') as file:
        file.write(tei)
    make_changelog(base_name)
    return (base_name, DictionaryStrategy.Imported)


def main():
    assert_correct_working_directory()
    if len(sys.argv) != 2:
        print("Error, path to shared FreeDict files required as first argument.")
        sys.exit(1)
    if not os.path.exists(sys.argv[1]):
        print("Error, path does not exist",sys.argv[1])
        sys.exit(2)
    # statistics
    too_small, manual = [], []
    api = get_fd_api()
    with multiprocessing.Pool(5) as p:
        res = p.starmap(import_dictionary, # ↓ pair with api, see import_dictionary
                ((api, l) for l in sorted(parse_links())))
        for action in res:
            dictname, action = action # what has been done with which dictionary
            if action is DictionaryStrategy.TooSmall:
                too_small.append(dictname)
            elif action == DictionaryStrategy.ManuallyEdited:
                manual.append(dictname)
    print("The following dictionaries were skipped because:")
    print("… a non-WikDict version exists:",', '.join(manual))
    print("… they were too small:", ', '.join(too_small))


if __name__ == '__main__':
    main()




"""This is a dummy file, so that this  directory can be used as a module. The
actual magic should *always* happen in tokenizer.py. Tests.py is a stand-alone
test suite and does, really, tests."""

from . import output, tokenizer
from . import output, output



#!/usr/bin/env python3
"""Releasing a bunch of dictionaries can be tedious because of the adjustments
to the header. This script automated the following:

* update the date tag
* update the copyright year within the availability information, because a
  new release also renews the copyright
* open $EDITOR to ask the user for a changelog message
* count headwords
* update the edition

For the concrete usage see the corresponding usage message."""

#pylint: disable=wrong-import-position
import datetime
import os
import re
import shutil
import sys

from fd_tool import config

def get_editor():
    """Detect an editor to use. Try to use $EDITOR or probe for a bunch of
    more common ones."""
    editor = None
    if 'EDITOR' in os.environ:
        editor = os.environ['EDITOR']
    if not editor:
        for name in ('vim', 'vi', 'nvim', 'emacs', 'nano', 'edit', 'notepad',
                'gedit', 'pluma', 'kate'): # a few more popular ones
            if shutil.which(name):
                editor = name
                break
    if not editor:
        print(("Failed to detect an editor. Please set the EDITOR environment "
                "variable to a program in your PATH."))
        sys.exit(1)
    else:
        return editor



class TagNotFoundException(Exception):
    pass

def find_tag(document, tag):
    """Find a tag in a document, returning start and end positions of the
    opening and closing tag. Raise TagNotFoundException if no such tag
    exists."""
    match = re.search(r'<\s*%s\b.*?>' % tag, document)
    if not match:
        raise TagNotFoundException(tag)
    opening_start, opening_end = match.span()
    # find closing tag
    match = re.search(r'<\s*/\s*%s\s*>' % tag, document[opening_end:])
    if not match:
        raise TagNotFoundException('no closing tag for `%s`' % tag)
    closing_start, closing_end = (c + opening_end for c in match.span())
    # python's ranges are exclusive, so add + 1
    return (opening_start, opening_end, closing_start, closing_end)

def get_text(document, tag):
    """Get the text of a specified tag. Raise TagNotFoundException if no such
    tag exists."""
    _, start, end, _ = find_tag(document, tag)
    return document[start:end]

def replace_tag_content(document, tag, new_text):
    """Insert the given new content into the first occurrence of this tag in the
    given document."""
    _, start, end, _ = find_tag(document, tag)
    return document[:start] + new_text + document[end:]

def get_user_info(conf):
    """Obtain (GitHub) user name and full name from configuration. Exit if user
    name is not set. Use system name if full name (AKA real name) is not set."""
    try:
        user_name = conf['DEFAULT']['user_name']
    except KeyError:
        sys.stderr.write(("user_name not set in FreeDict configuration. Please "
                "set it and try again."))
        sys.exit(28)
    try:
        real_name = conf['DEFAULT']['full_name']
    except KeyError:
        # guess name, could be not correct
        if 'win32' in sys.platform or 'wind' in sys.platform:
            import getpass
            real_name = getpass.getuser()
        else: # on unixoids, use pwd
            import pwd
            real_name = pwd.getpwuid(os.getuid())[4]
            # on some systems, real name end with commas, strip those
            while real_name and not real_name[-1].isalpha():
                real_name = real_name[:-1]
    return (user_name, real_name)


def add_changelog_entry(document, edition, date, username, author):
    """Try to detect an text editor, open it and add the written content to a
    new change tag within the supplied revision_desc."""
    editor = get_editor()
    fn = 'changelog.tmp'
    with open(fn, 'w', encoding='UTF-8') as f:
        f.write("""\n
# Please enter your change notes as you would enter them in a </change> tag in
# the revisionDesc tag of a TEI header. Empty lines will be ignored and lines
# starting with a hash `#` will be ignored.
# Valid formatting include either plain text or lists like
# `<list><item>blah</item></list>`. Please note that you need to take care of
# escaping yourself.
""")
    ret = os.system('%s %s' % (editor, fn))
    if ret:
        print("Error while starting", editor)
        sys.exit(2)
    with open(fn, encoding='UTF-8') as f:
        data = '\n'.join(l.rstrip() for l in f
                if l.strip() and not l.lstrip().startswith('#'))
    os.remove(fn)
    if not data.strip():
        print("No changes, aborting…")
        sys.exit(0)
    # insert full name if no reference to the user name found, otherwise use the
    # reference
    change = None
    if re.search('id\\s*=\\s*(?:\'|")#' + username, document):
        change = '<change n="{}" when="{}" who="#{}">\n'.format(edition, date,
            username.rplace('#', ''))
    else:
        change = '<change n="{}" when="{}">\n<name>{}</name>\n' \
                .format(edition, date, author)
    change += '%s\n</change>' % data
    latest_change, _, _, _ = find_tag(document, 'change')
    latest_change_tag = latest_change
    if latest_change > 0 and document[latest_change-1].isspace():
        latest_change -= 1
        while latest_change > 0 and document[latest_change].isspace() and \
                document[latest_change] != '\n':
            latest_change -= 1
    indent = document[latest_change+1:latest_change_tag]
    change = change.rstrip().replace('\n', '\n' + indent)
    return ''.join((document[:latest_change],
                '\n', indent, change, '\n', indent,
                document[latest_change_tag:])).lstrip()

def update_date(document, date):
    """Find publicationStmt/date, update it."""
    try:
        opening_start, _, _, closing_end = find_tag(document, 'date')
    except TagNotFoundException:
        sys.stderr.write(("Warning: <date> tag not found. It's advised that "
            "this is added to the header.\n"))
        return document
    if opening_start < 0:
        return document # no date, no action
    # try to detect whether this date is within a change tag
    change = document[:opening_start].rfind('<change')
    if change > 0:
        return document
    date = datetime.datetime.now().strftime('<date when="%Y-%m-%d">%b %d, %Y</date>')
    return document[:opening_start] + date + document[closing_end:]

def update_edition(document, version):
    """Update edition within the edition tag."""
    return replace_tag_content(document, 'edition', version)

def update_extent(document):
    """Count headwords within document and update the headword count of the
    extend tag."""
    headwordcount = len(re.findall(r'<\s*entry.*?>', document))
    return replace_tag_content(document, 'extent', '%s headwords' % headwordcount)


def update_copyright(document):
    """Find a stanza containing "(c) 2014-2017 xyz" or "© 2020 foo" to update
    the year and therfore update copyright information."""
    availability = get_text(document, 'availability')
    match = re.search(r'(©|\([cC]\))\s*([0-9]{4})(?:-)([0-9]{4})', availability)
    if not match:
        return document
    start, end = match.span()
    year = match.groups()[1]
    if match.groups()[2]:
        year = match.groups()[2]
    availability = availability[:start] + availability[start:end].replace(year,
                    datetime.datetime.now().strftime('%Y')) + \
                availability[end:]
    return replace_tag_content(document, 'availability', availability)

def parse_args():
    def usage(msg=None):
        if msg:
            print(msg)
        print("Usage: make E=<EDITION>")
        print("   or: %s <EDITION> <INPUT_FILE>" % \
                os.path.basename(sys.argv[0]))
        print("\nThis script assists in changing a TEI header for release. It does the following:")
        print("- update the edition")
        print("- updates the release date")
        print("- adds a new changelog entry")
        print("- update the year in the availability / copyright section, if appropriate")
        print("\nExample: make E=1.5.2 lat-deu.tei")
        sys.exit(0)
    if len(sys.argv) == 2 and sys.argv[1] in ('-h', '--help'):
        usage()
    elif len(sys.argv) != 3:
        usage("The arguments edition and input file are mandatory.")
    else:
        return sys.argv[1:3]

def main():
    edition, input_file = parse_args()
    conf = config.discover_and_load()
    isodate = datetime.datetime.today().strftime('%Y-%m-%d')
    username, full_name = get_user_info(conf)
    with open(input_file, 'r', encoding='UTF-8') as f:
        document = f.read()
    document = add_changelog_entry(document, edition, isodate, username,
            full_name)
    document = update_date(document, isodate)
    document = update_copyright(document)
    document = update_extent(document)
    document = update_edition(document, edition)
    with open(input_file, 'w', encoding='UTF-8') as f:
        f.write(document)
    print("You might want to re-indent the latest <change/> to fit it to the rest of the document.")

if __name__ == '__main__':
    main()



#pylint: disable=too-many-public-methods,import-error,too-few-public-methods,missing-docstring,unused-variable,multiple-imports
import unittest
import tokenizer as tk
from tokenizer import ChunkType

class tests(unittest.TestCase):
    def test_commas_are_detected(self):
        chunks = tk.tokenize('foo, bar')
        self.assertEqual(len(chunks), 3)
        self.assertEqual(chunks[0][0], ChunkType.Word)
        self.assertEqual(chunks[1][0], ChunkType.Comma)
        self.assertEqual(chunks[2][0], ChunkType.Word)

    def test_semicolons_are_detected(self):
        chunks = tk.tokenize('foo; bar')
        self.assertEqual(len(chunks), 3)
        self.assertEqual(chunks[0][0], ChunkType.Word)
        self.assertEqual(chunks[1][0], ChunkType.Semicolon)
        self.assertEqual(chunks[2][0], ChunkType.Word)

    def test_vert_bar_detected(self):
        chunks = tk.tokenize('foo | bar')
        self.assertEqual(len(chunks), 3)
        self.assertEqual(chunks[0][0], ChunkType.Word)
        self.assertEqual(chunks[1][0], ChunkType.VerticalBar)
        self.assertEqual(chunks[2][0], ChunkType.Word)

    def test_parenthesized_expressions_parsed_corretly(self):
        tks = tk.tokenize('test (the) behaviour')
        self.assertEqual(len(tks), 3)
        self.assertEqual(tks[0][0], ChunkType.Word)
        self.assertEqual(tks[1][0], ChunkType.Paren)
        self.assertEqual(tks[1][1], 'the')
        self.assertEqual(tks[2][0], ChunkType.Word)
        # now paren at beginning
        tks = tk.tokenize('(the) behaviour')
        self.assertEqual(len(tks), 2)
        self.assertEqual(tks[0][0], ChunkType.Paren)
        self.assertEqual(tks[0][1], 'the')
        self.assertEqual(tks[1][0], ChunkType.Word)
        # now at the end
        tks = tk.tokenize('fond (of)')
        self.assertEqual(len(tks), 2)
        self.assertEqual(tks[0][0], ChunkType.Word)
        self.assertEqual(tks[1][0], ChunkType.Paren)
        self.assertEqual(tks[1][1], 'of')

    def test_braces_recognized(self):
        tks = tk.tokenize('test {blah} behaviour')
        self.assertEqual(len(tks), 3)
        self.assertEqual(tks[0][0], ChunkType.Word)
        self.assertEqual(tks[1][0], ChunkType.Brace)
        self.assertEqual(tks[1][1], 'blah')
        self.assertEqual(tks[2][0], ChunkType.Word)
        # now paren at beginning
        tks = tk.tokenize('{schnurp} xyz')
        self.assertEqual(len(tks), 2)
        self.assertEqual(tks[0][0], ChunkType.Brace)
        self.assertEqual(tks[1][0], ChunkType.Word)
        self.assertEqual(tks[0][1], 'schnurp')
        # now at the end
        tks = tk.tokenize('Fleisch {n}')
        self.assertEqual(len(tks), 2)
        self.assertEqual(tks[0][0], ChunkType.Word)
        self.assertEqual(tks[1][0], ChunkType.Brace)
        self.assertEqual(tks[1][1], 'n')

    def test_brackets_recognized(self):
        tks = tk.tokenize('test [Br.] behaviour')
        self.assertEqual(len(tks), 3)
        self.assertEqual(tks[0][0], ChunkType.Word)
        self.assertEqual(tks[1][0], ChunkType.Bracket)
        self.assertEqual(tks[1][1], 'Br.')
        self.assertEqual(tks[2][0], ChunkType.Word)
        # now paren at beginning
        tks = tk.tokenize('[zool.] xyz')
        self.assertEqual(len(tks), 2)
        self.assertEqual(tks[0][0], ChunkType.Bracket)
        self.assertEqual(tks[1][0], ChunkType.Word)
        self.assertEqual(tks[0][1], 'zool.')
        # now at the end
        tks = tk.tokenize('Fleisch [bot.]')
        self.assertEqual(len(tks), 2)
        self.assertEqual(tks[0][0], ChunkType.Word)
        self.assertEqual(tks[1][0], ChunkType.Bracket)
        self.assertEqual(tks[1][1], 'bot.')

    def test_enclosing_chars_have_precedence_over_delimiters(self):
        tks = tk.tokenize('(a,b;c|d/e) {a,b;c|d/e} [a,b;c|d/e]')
        self.assertEqual(len(tks), 3)
        for chunk in tks:
            self.assertEqual(chunk[1], 'a,b;c|d/e')
        # test slash
        tks = tk.tokenize('x /a,b;c|d/ y', parse_slash=True)
        self.assertEqual(len(tks), 3)
        self.assertEqual(tks[1][1], 'a,b;c|d')

    def test_paren_or_slashes_in_words_ignored(self):
        tks = tk.tokenize('house(s), Häuser/innen', parse_slash=True)
        self.assertEqual(len(tks), 3)
        self.assertEqual(tks[0][0], ChunkType.Word)
        self.assertEqual(tks[2][0], ChunkType.Word)

    def test_slash_within_paren_works(self):
        tks = tk.tokenize('(my/example)')
        self.assertEqual(len(tks), 1)
        self.assertEqual(tks[0][0], ChunkType.Paren)
        tks = tk.tokenize('(my/example)', parse_slash=True)
        self.assertEqual(len(tks), 1, "expected one element, got: " + repr(tks))
        self.assertEqual(tks[0][0], ChunkType.Paren)

    def test_only_expressions_with_no_spaces_withing_slash_slash_parsed(self):
        print("===")
        tks = tk.tokenize('/AB/', parse_slash=True)
        print("---")
        self.assertEqual(len(tks), 1)
        self.assertEqual(tks[0][0], ChunkType.Slash)
        tks = tk.tokenize('A / B/', parse_slash=True)
        self.assertEqual(len(tks), 1)
        self.assertEqual(tks[0][0], ChunkType.Word)
        
     
 

if __name__ == '__main__':
    unittest.main()



"""Collect all information about released dictionaries."""

from datetime import datetime
import distutils.version
import hashlib
import json
import os
import re
import shutil
import subprocess
import sys
import urllib.request

import semver

from .config import RELEASE_HTTP_TOOL_BASE
from .dictionary import DownloadFormat, normalize_version

# relative to github.com/freedict/
TOOLS_REPO = 'tools'
USER_AGENT = 'Friendly FreeDict Helper'

class ReleaseError(Exception):
    """This error can occur, when information about a release is gathered. It
    wraps all sorts of ValueErrors and IoErrors."""


def git(cmd):
    """Execute a git command with the given list as arguments. Return stdout."""
    proc = subprocess.Popen(['git'] + cmd, cwd=os.environ['FREEDICT_TOOLS'],
            stdout=subprocess.PIPE)
    stdout = proc.communicate()[0].strip()
    ret = proc.wait()
    if ret:
        raise ReleaseError("`git %s` exited with exit code %d" % \
                (' '.join(cmd), ret))
    return stdout.decode(sys.getdefaultencoding())

def get_tools_release():
    """Retrieve the latest FreeDicttools release as a tuple with containing
    (version, date, downloadlink)."""
    if not 'FREEDICT_TOOLS' in os.environ or not shutil.which('git'):
        raise ReleaseError(("Unable to retrieve list of releases of "
            "FreeDict tools. Either FREEDICT_TOOLS is unset or git not "
            "installed."))
    releases = git(['tag']).split('\n')
    max_ver = '0.0.0'
    for tag in releases:
        max_ver = semver.max_ver(max_ver, tag)
    if max_ver == '0.0.0':
        raise ReleaseError("No tools releases found.")
    date = re.search('^([0-9]{4}-[0-9]{2}-[0-9]{2})',
            git(['show', '-s', '--pretty=format:%ci'])).groups()[0]
    return (max_ver, date, '{}/freedict-tools-{}.tar.xz'.format(
            RELEASE_HTTP_TOOL_BASE, max_ver))

def get_release_info_for_dict(path, version):
    """Retrieve information about the releases of a dictionary."""
    files = {}
    name = None
    version = normalize_version(version)
    for file in os.listdir(path):
        format = DownloadFormat.get_type(file)
        if not format:
            continue # ignore unknown file naming, possibly outdated or unsupported formats

        parsed_name, file_version_str = format.value.search(file).groups()
        file_version = normalize_version(file_version_str)

        if file_version != version:
            raise ReleaseError('Version from file name "%s" did not match version of directory "%s"' \
                    % (file, version))
        if not name:
            name = parsed_name
        elif parsed_name != name:
            raise ReleaseError('Found two different dictionary identifiers in %s: %s and %s' \
                        % (path, parsed_name, name))
        full_path = os.path.join(path, file)
        try:
            with open(full_path + '.sha512', 'r') as f:
                sha = f.read().strip().split('  ')[0]
        except FileNotFoundError:
            raise FileNotFoundError('expected a sha512 checksum, found nothing',
                    full_path + '.sha512', 'r')
        files[full_path] = (name, format, sha)
    if not files:
        raise ReleaseError("no downloads available for %s" % path)
    return files


def get_all_downloads(root):
    """Get all paths to a downloadable file from a given root.
    This function walks the given path and returns all paths to files with the
    relative to the given root."""
    # only match directories with two iso 639-3 codes, separated by "-"
    dirpattern = re.compile(r'[a-z]{3}-[a-z]{3}')
    release_directories = tuple(os.path.join(root, e) for e in os.listdir(root)
            if os.path.isdir(os.path.join(root, e)) and dirpattern.search(e) \
                    and not 'tools' in e.lower())
    if not release_directories:
        raise ReleaseError("%s: no released dictionary found" % root)

    dictionaries_with_release = {} # collect dictionaries which have a release
    for dictdir in release_directories:
        # iterate over subdirectories which represent versions
        for version_dir in (os.path.join(dictdir, f) for f in os.listdir(dictdir)):
            version = os.path.basename(version_dir)

            try:
                files = get_release_info_for_dict(version_dir, version)
            except ReleaseError as e:
                if e.args[0].startswith('no downloads available'):
                    continue # skip versions with no or unknown releases
                else:
                    raise
            if not files:
                print('Warning: no files in "%s", skipping...' % version)
                continue
            name = next(iter(files.values()))[0]
            if not name in dictionaries_with_release:
                dictionaries_with_release[name] = {}
            # transform {path: (name, format, hash)} to (path, format, hash)
            dictionaries_with_release[name][version] = tuple((k, v[1], v[2])
                    for k,v in files.items())
    return dictionaries_with_release


def get_latest_version(release_information):
    """Iterate over object and return the latest version, as defined by
    semver. Versions that are valid distutils.version.StrictVersions, but not
    valid semver versions are converted."""
    latest = None
    latest_strict = None # might contain '-' replaced through '.'
    for version in release_information:
        version_strict = version[:]
        try:
            normalize_version(version_strict)
        except ValueError as e:
            raise ReleaseError(e.args)
        if not latest:
            latest = version
            latest_strict = version_strict
        else:
            if version_strict > latest_strict:
                latest = version
                latest_strict = version_strict
    if not latest:
        raise ReleaseError("No versions found for " % repr(release_information))
    return latest

def github_request(path):
    """Make a GitHub API request and return aJSON object."""
    url = "https://api.github.com/{}".format(path)
    request_headers = {'User-Agent': USER_AGENT}
    request = urllib.request.Request(url, headers=request_headers)
    with urllib.request.urlopen(request) as f:
        return json.loads(f.read().decode('UTF-8'))

def get_latest_tools_release():
    latest = {'name': '0.0.0'}
    for tag in github_request("repos/freedict/{}/tags".format(TOOLS_REPO)):
        latest = max(latest, tag, key=lambda t: t['name'])
    if latest['name'] == '0.0.0':
        raise ValueError("could not find a release for FreeDict tools")
    commit_url = latest['commit']['url']
    api_suffix = commit_url[commit_url.find('.com/') + 5:]
    commit_meta = github_request(api_suffix)['commit']['committer']
    # validate format
    date = datetime.strptime(commit_meta['date'].split('T')[0], '%Y-%m-%d')
    date = date.strftime('%Y-%m-%d')
    # We don't use the API-provided download URL (latest['tarball']), as it
    # contains a funny number in the unpacked directory.
    github_uri = f"https://github.com/freedict/tools/archive/refs/tags/{latest['name']}.tar.gz"
    request = urllib.request.Request(github_uri,
            headers={'User-Agent': USER_AGENT})
    with urllib.request.urlopen(request) as f:
        checksum = hashlib.sha512(f.read()).hexdigest()
    return {'version': latest['name'],
        'date': date,
        'URL': github_uri,
        'checksum': checksum}



"""Tests for fd_tool.api.dictionary."""

#pylint: disable=too-many-public-methods,import-error,too-few-public-methods,missing-docstring,unused-variable,multiple-imports
import unittest


from fd_tool.api.dictionary import Dictionary

class TestTictionary(unittest.TestCase):
    def test_unset_mandatory_fields_are_detected(self):
        # test for `date`
        d = Dictionary('lat-deu')
        d['headwords'] = 80
        d['edition'] = '0.1.2'
        self.assertFalse(d.is_complete())

    def test_is_complete_recognizes_all_mandatory_fields(self):
        # test for `date`
        d = Dictionary('lat-deu')
        d['headwords'] = 80
        d['edition'] = '0.1.2'
        d['date'] = '1871-12-13'
        self.assertTrue(d.is_complete())




"""FreeDict API generator

This script parses the meta data of all available dictionaries, queries
information about available releases and writes all information to an XML file.
For more information, please have a look at the wiki at
<http://freedict.org/howto>.

For usage of the script, try the -h option.
"""
import argparse
import os
import sys
import time

from fd_tool.api import dictionary, jsonhandlers, metadata, releases, xmlhandlers
from fd_tool import config
from fd_tool.config import get_path



def exec_or_fail(command):
    """If command is not None, execute command. Exit upon failure."""
    if command:
        ret = os.system(command)
        if ret:
            print("Failed to execute `%s`:" % command)
            sys.exit(ret)


def read_dict_info(conf, generate_api=True):
    """Parse dictionary meta data from the dictionary source and extract
    information about all released dictionaries. Return a list of dictionaries
    (Dictionary objects). generate_api is used to emit a message which is only
    useful if the API is generated."""
    dictionaries = []
    for dict_source in ['crafted', 'generated']:
        dict_source = get_path(conf[dict_source])
        print("Parsing meta data for all dictionaries from", dict_source)
        dictionaries.extend(metadata.get_meta_from_xml(dict_source))

    release_path = get_path(conf['release'])
    print("Parsing release information from", release_path)
    release_files = releases.get_all_downloads(release_path)
    for dict in dictionaries:
        name = dict.get_name()
        if not name in release_files:
            if generate_api:
                print("Skipping %s, no releases found." % name)
            continue
        try:
            version = releases.get_latest_version(release_files[name])
        except releases.ReleaseError as e: # add file name to error
            raise releases.ReleaseError(list(e.args) + [name])
        for full_file, format, sha in release_files[name][version]:
            dict.add_download(dictionary.mklink(full_file, format, version, sha))
    return dictionaries



def find_outdated_releases(dictionaries):
    """This function finds dictionaries which have been updated, but not
    released yet."""
    candidates = []
    for dict in dictionaries:
        if dict.get_downloads() == []:
            candidates.append((dict.get_name(), None, None))
        else:
            released = max(l.version for l in dict.get_downloads())
            if dict['edition'] > released:
                candidates.append((dict.get_name(), dict['edition'], released))
    return candidates


def main_body(args):
    parser = argparse.ArgumentParser(description='FreeDict API generator')
    parser.add_argument("-n", "--need-update", dest="check_for_unreleased_dicts",
            action="store_true", default=False,
            help="check for unreleased dictionaries instead of generating the API file")
    parser.add_argument('-p', "--pre-exec-script", dest="prexec", metavar="PATH",
            help=('script/command to execute before this script, e.g. to set up a sshfs '
                'connection to a remote server, or to invoke unison.'))
    parser.add_argument('-o', "--post-exec-script", dest="postexc", metavar="PATH",
            help=("script/command to execute after this script is done, e.g. to "
                "umount mounted volumes."))

    args = parser.parse_args(args[1:])
    conf = config.discover_and_load()

    exec_or_fail(args.prexec) # mount / synchronize release files

    dictionaries = read_dict_info(conf, not args.check_for_unreleased_dicts)

    if args.check_for_unreleased_dicts:
        outdated = find_outdated_releases(dictionaries)
        if not outdated:
            print("Everything up-to-date.")
        else:
            print("\nName      Source Version    Release Version")
            print("-------   ---------------   --------------------------")
            for data in sorted(outdated, key=lambda x: x[0]):
                name, v1, v2 = [str(e if e else 'unknown') for e in data]
                print('{}   {:<15}   {:<15}'.format(name, v1, v2))
    else:
        # remove dictionaries without download links
        dictionaries = sorted((d for d in dictionaries if d.get_downloads() != []),
            key=lambda entry: entry.get_name())
        tools = releases.get_latest_tools_release()
        api_path = config.get_path(conf['DEFAULT'], key='api_output_path')
        xml_path = os.path.join(api_path, 'freedict-database.xml')
        json_path = os.path.join(api_path, 'freedict-database.json')
        if not os.path.exists(api_path):
            os.makedirs(os.path.dirname(api_path))
        print("Writing XML API file to",xml_path)
        xmlhandlers.write_freedict_database(xml_path, dictionaries, tools)
        print("Writing JSON API file to",json_path)
        jsonhandlers.write_freedict_database(json_path, dictionaries, tools)

    # if the files had been mounted with sshfs, it's a good idea to give it some
    # time to synchronize its state, otherwise umounting fails
    time.sleep(2)
    exec_or_fail(args.postexc) # umount or unison files, if required

def main():
    """Wrapper for nicer error case handling."""
    #pylint: disable=broad-except
    try:
        main_body(sys.argv)
    except Exception as e:
        if 'DEBUG' not in os.environ:
            print('Error:',str(e))
            print(("\nNote: Rerun the script with the environment variable DEBUG=1 "
                "to obtain a traceback."))
            sys.exit(9)
        else:
            raise e

if __name__ == '__main__':
    main()



"""This file offers common configuration parsing facilities required for the
Freedict build system."""

import configparser
import os
import re

class ConfigurationError(Exception):
    def __init__(self, msg, path=None):
        super().__init__()
        self.path = path
        self.msg = msg
    def __repr__(self):
        if self.path:
            return 'error in configuration "%s": %s' % (self.path, self.msg)
        else:
            return self.msg

    def __str__(self):
        return repr(self)



def load_configuration(conffile):
    """Load given `config` from given path. Default values are provided and
    missing mandatory options will raise a ConfigurationError."""
    config = configparser.ConfigParser()
    config['DEFAULT'] = {
            'file_access_via': 'sshfs', # unison or sshfs possible
            'api_output_path': '.'}
    config['crafted'] = {}
    config['crafted']['local_path'] = ''
    config['generated'] = {}
    config['generated']['server'] = 'freedict.org'
    config['generated']['remote_path'] = '/var/www/download/generated'
    config['generated']['local_path'] = ''
    config['generated']['user'] = 'anonymous'
    config['generated']['skip'] = 'no'
    config['release'] = {}
    config['release']['server'] = 'freedict.org'
    config['release']['remote_path'] = '/var/www/download/dictionaries'
    config['release']['local_path'] = ''
    config['release']['user'] = 'anonymous'
    config['release']['skip'] = 'no'

    # overwrite defaults with user settings
    with open(conffile) as configfile:
        config.read_file(configfile)

    if config['DEFAULT']['file_access_via'] not in ['sshfs', 'unison']:
        raise ConfigurationError(('section=DEFAULT, file_access_via="%s": '
            'invalid value, possible values are sshfs and unison') \
                % config['DEFAULT']['file_access_via'], conffile)
    api_path = config['DEFAULT']['api_output_path']
    if api_path and os.path.exists(api_path) and os.path.isfile(api_path):
        raise ConfigurationError("expected directory, found file", path=api_path)

    for section in ('generated', 'crafted', 'release'):
        if not config[section]['local_path']:
            raise ConfigurationError("error, local_path not set for section [%s]" \
                    % section, conffile)
        path = get_path(config[section])
        if not os.path.exists(path):
            raise ConfigurationError("path \"%s\" configured in section [%s] doesn't exist" \
                    % (path, section), conffile)

    return config

def discover_and_load():
    """This file attempts to discover and load a FreeDict configuration file at
    the usual places. If no configuration was found, a ConfigurationError is
    raised."""
    paths = [os.path.join(os.path.expanduser("~"), '.config/freedict/freedictrc')]
    if os.environ.get('LOCALAPPDATA'):
        paths.append(os.path.join(os.environ['LOCALAPPDATA'], 'freedict/freedict.ini'))
    conffile = [path for path in paths if os.path.exists(path)]
    if not conffile:
        phrase = ('one of the following directories' if len(paths) > 1 else 'the following directory')
        raise ConfigurationError(("no configuration found. Please initialize "
            "one in " + phrase + ' ' + ', '.join(paths)))
    return load_configuration(conffile[0])


def get_path(section, key='local_path'):
    """Return a local_path from a section with $HOME, ~/ or %HOME% replaced."""
    home = os.path.expanduser('~')
    path = section[key]
    for pattern in (r'\$HOME', '~', '%HOME%'):
        path = re.sub('^' + pattern, home, path)
    return path





"""This script makes remote files available for local processing. Remote files
are e.g. the released files hosted on a server as downloads or the
auto-generated dictionaries, kept outside the git repository.
This script requires a configuration. Please see the README for more details.
Running this script with the `-h` option will give an overview about its usage."""

import argparse
import os
import subprocess
import sys


from .. import config


def execute(cmd, raise_on_error=False):
    """Execute a command; if the return value is != 0, the program either
    terminates or an exception is raised."""
    proc = subprocess.Popen(cmd, shell=True, stdout=subprocess.PIPE,
            stderr=subprocess.PIPE)
    text = (e.decode(sys.getdefaultencoding()) for e in proc.communicate())
    ret = proc.wait()
    if ret:
        text = '\n'.join(text).strip()
        if text.startswith('fusermount: ') and 'not found in /etc/mtab' in text:
            return # umounting something which isn't mounted is not harmful, ignore

        text = ('Subcommand failed with exit code %s\n'
                 'Command: %s\n%s\n') % (ret, cmd, text)
        if raise_on_error:
            raise OSError(text)
        else:
            print(text)
            if ret >= 255:
                ret = 1
            sys.exit(ret)

class UnisonFileAccess:
    """This class is one of two classes to allow access to remote files using
    unison. The drawback with unison is that before the usage by other scripts,
    all files have to be downloaded. On the other hand, this might speed up
    subsequent runs and allows offline work. On Windows, it might be desirable
    to use unison, because sshfs is not officially ported to Windows."""
    def __init__(self):
        # save make_available arguments for make_unavailable
        self.args = tuple()

    def name(self):
        return "unison"

    def make_available(self, user, server, remote_path, path):
        """Synchronize files to have them available locally."""
        self.args = (user, server, remote_path, path)
        # set UNISON=`path` to avoid usage of any $HOME/.unison/default.prf
        oldunison = None
        if 'UNISON' in os.environ:
            oldunison = os.environ['UNISON']
        os.environ['UNISON'] = os.path.join(path, '.unison')
        ret = os.system("unison -auto -log -times -contactquietly -terse " + \
                "-ignore 'Regex .*.swp' -ignore 'Regex .*.swo' " + \
                "-ignore 'Regex .*/build' " + \
                "-ignore 'Regex .*~' -ignore 'Regex .unison.*' " + \
                "ssh://{}@{}/{}/ {}".format(user, server,
                    remote_path, path))
        if ret:
            raise OSError("Process gave error code %d" % ret)
        if oldunison:
            os.environ['UNISON'] = oldunison

    #pylint: disable=unused-argument
    def make_unavailable(self, path):
        """Synchronise in case files were created."""
        if not self.args:
            return
        self.make_available(*self.args)

class SshfsAccess:
    """This class mounts and umounts the remote files using sshfs. This will
    work on any system that fuse runs on, namely GNU/Linux, FreeBSD and Mac."""
    def name(self):
        return 'sshfs'

    def make_available(self, user, server, remote_path, path):
        """Mount remote file system using sshfs."""
        # is mounted?
        if os.path.ismount(path): # mounted, -m help says we need to return 201
            return 201 # and no action
        if len(os.listdir(path)) > 0:
            print("Error: %s has to be empty, otherwise mounting impossible." % path)
            sys.exit(41)
        return execute('sshfs {}@{}:{} {}'.format(user, server, remote_path, path))

    def make_unavailable(self, path):
        execute('fusermount -u {}'.format(path), raise_on_error=True)



def setup():
    """Find freedict directory and parse command line arguments. Return a tuple
    with the freedict directory and the configuration object."""
    # parse command line options
    parser = argparse.ArgumentParser(description='FreeDict build setup utility')
    parser.add_argument('-a', dest="print_api_path", action='store_true',
            help=("print output directory where the freedict-database.xml and "
                "json are stored;  the value is read from the local "
                "configuration"))
    parser.add_argument('-m', dest="make_available", action='store_true',
            help='''make files in generated/ and release/ available; this will \
                    use internally either sshfs or unison, depending on the \
                    configuration. The script will exit with 0 on success, \
                    with 201 if remote filesystem was mounted already and with \
                    the error code of the appropriate subcommand otherwise.''')
    parser.add_argument('-r', dest="print_release_path", action='store_true',
            default=False, help=("print output directory to which releases are "
                "deployed. the value is read from the local configuration"))
    parser.add_argument('-u', dest='umount', action='store_true',
        help='clean up actions for release/ and generated/, e.g. umount of fuse mount points, etc.')
    args = parser.parse_args()

    # check for contradicting options
    if args.umount and args.make_available:
        print("Error: you can only specify -u or -m exclusively.")
        sys.exit(44)
    if not any((args.umount, args.make_available, args.print_api_path,
            args.print_release_path)):
        print("Error: No option specified")
        parser.print_help()
    return args


def main():
    args = setup()
    try: # load configuration
        conf = config.discover_and_load()
    except config.ConfigurationError as e:
        print(e)
        sys.exit(42)

    if args.print_api_path:
        print(config.get_path(conf['DEFAULT'],
            key='api_output_path'))
        sys.exit(0)
    elif args.print_release_path:
        print(config.get_path(conf['release'],
            key='local_path'))
        sys.exit(0)

    access_method = UnisonFileAccess()
    if conf['DEFAULT']['file_access_via'] == 'sshfs':
        access_method = SshfsAccess()

    release_directory = config.get_path(conf['release'])
    if not os.path.exists(release_directory):
        try:
            os.makedirs(release_directory)
        except OSError:
            # if the file does exist, but the fuse endpoint is _not_ connected,
            # we could try running fusermount -u:
            os.system('fusermount -u "%s"' % release_directory)

    ret = 0
    if args.make_available:
        for section in ('release', 'generated'):
            if conf[section].getboolean('skip'):
                print("Skipping",section)
                continue
            print('Making files for "%s" available...' % section)
            options = conf[section]
            target_path = config.get_path(options)
            ret = access_method.make_available(options['user'], options['server'],
                options['remote_path'], target_path)
    elif args.umount:
        for section in ('generated', 'release'):
            if conf[section].getboolean('skip'):
                print("Skipping",section)
                continue
            target_path = config.get_path(conf[section])
            try:
                access_method.make_unavailable(target_path)
            except OSError as e:
                print(e.args[0])
                continue
    sys.exit(ret)

if __name__ == '__main__':
    main()



"""This file provides classes to parse all dictionaries to extract the required
meta data for the freedict-database.xml. The meta data comes from the header of
each dictionary and gives details on the quality or size of a dictionary."""
import datetime
import html.parser
import io
import os
import re
from xml.etree import ElementTree as ET

from . import dictionary
from . import xmlhandlers
from .xmlhandlers import istag

class MetaDataParser(xmlhandlers.TeiHeadParser):
    """Parse a TEI XML dictionary for required and optional meta data using
    Python's iterparse facility. It guesses the path on it's own, unless a path
    is given as optional parameter.  For each element where data should be
    extracted, there is a handle_<tagname> which is called if it exists.
    This is a common class from which concrete parserfs can be derived. They all
    share the facility to parse TEI XML, but they can differ in they way they
    retrieve the TEI XML. While the local file system is a natural choice,
    one could choose to retrieve the information over HTTP(s).
    After the header has been parsed, the .dictionary attribute holds a reference
    to a populated Dictionary object."""
    # pattern to match mtaintainer in <resp/> tag
    MAINTAINER_PATTERN = re.compile(r'^([^<]+)\s*<?([^<]+)?>$')
    # match number of headwords in <extend/> pattern
    HEADWORD_PATTERN = re.compile(r'(\d+(?:\.|,|\s*)\d*).*')
    # ISO date pattern
    DATE_PATTERN = re.compile(r'\d{4}-\d{2}-\d{2}')

    def __init__(self, name, xml):
        # can be a file object or a string with XML data
        xml = self.get_file_object(xml)
        super().__init__(xml)
        self.dictionary = dictionary.Dictionary(name)
        self.dictionary['date'] = None

    def get_file_object(self, src):
        """Transparently create a file object. If this method gets a file
        object, it will return is unchanged, otherwise the given string is
        wrapped in a StringIO object."""
        if hasattr(src, 'read') and hasattr(src, 'close'):
            return src # is already file object
        elif isinstance(src, bytes):
            return io.BytesIO(src)
        elif isinstance(src, str):
            return io.StringIO(src)
        else:
            raise ValueError("Either file object, str or byte array " + \
                    "expected, got %s." % type(src))

    def parse(self):
        try:
            super().parse()
        except ET.ParseError as e:
            print(("Warning: while parsing {} an error occurred. Might be still "
                    "ok though.{}").format(self.dictionary.get_name(),
                        '; '.join(e.args)))

        # check whether parsing was successful
        if not self.dictionary.is_complete():
            missing = [k for k in self.dictionary.get_mandatory_keys()
                        if not self.dictionary[k]]
            raise ValueError("%s: the following information couldn't be read: "\
                % self.dictionary.get_name() + ', '.join(missing))

    def handle_tag(self, elem):
        """Delegate parsing of XML tags to specialised functions."""
        if istag(elem, 'date') or istag(elem, 'change'):
            self.__extract_date(elem)
        tag = elem.tag.split(self._namespace)[-1] # strip etree namespace
        funcname = 'handle_%s' % tag
        if not hasattr(self, funcname):
            return
        result = getattr(self, funcname)(elem)
        if result:
            self.dictionary.update(result)


    def handle_sourceDesc(self, node):
        """Extract a source url, if any."""
        ptr = node.findall('.//%sptr' % self._namespace)
        if ptr:
            return {'sourceURL': ptr[0].get('target')}
        # search for source URL within '<ref target=""')
        ref = node.findall('.//%sref' % self._namespace)
        if ref:
            return {'sourceURL': ref[0].get('target')}
        return None


    def handle_edition(self, elem):
        return {'edition': elem.text[:]}


    def handle_notesStmt(self, node):
        """Try to extract status."""
        for note in node.findall(self._namespace + 'note'):
            if note.get('type') == 'status':
                return {'status': note.text}
        raise ValueError('A notesStmt without a <note type="status"/> encountered.')



    def handle_extent(self, elem):
        """Extract extent (number of headwords)."""
        match = MetaDataParser.HEADWORD_PATTERN.search(elem.text)
        if not match:
            raise ValueError("Could not extract number of headwords from " +
                    repr(elem.text))
        headwords = ''.join(char for char in match.groups()[0] if char.isdigit())
        return {'headwords': headwords}


    def handle_respStmt(self, respStmt):
        """Maintainer is in <respStmt/>, this one can contain nesting for author
        and maintainer, hide complexity and return either maintainer, then
        author or None."""
        # find name attribute
        name = respStmt.find(self._namespace + 'name')
        if name is None:
            return
        resp = respStmt.findall(self._namespace + 'resp')
        if resp is None or not any('maintainer' in t.text.lower() for t in resp):
            return
        maintainer = name.text
        if not maintainer:
            return
        if 'up for grab' in maintainer.lower(): # not a real maintainer
            return
        # try to extract email address:
        if '@' in maintainer:
            maintainer = html.parser.unescape(maintainer)
        match = self.MAINTAINER_PATTERN.search(maintainer)
        if match:
            return {'maintainerName': match.groups()[0].rstrip().lstrip(),
                    'maintainerEmail': match.groups()[1].rstrip().lstrip()
                   }
        else:
            return {'maintainerName': maintainer.rstrip().lstrip()}

    def __extract_date(self, elem):
        """If date has not been set with the <date/> attrbiute in the header,
        guess it from change log."""
        parse_date = lambda x: datetime.datetime.strptime(x, '%Y-%m-%d')
        def set_date(isodate):
            try:
                if self.dictionary['date'] is None:
                    self.dictionary['date'] = isodate
                elif parse_date(isodate) > parse_date(self.dictionary['date']):
                    self.dictionary['date'] = isodate
            except ValueError:
                pass # we cannot parse all date formats
        if istag(elem, 'date'):
            if elem.get('when'): # use when, is usually in desired format
                set_date(elem.get('when').strip())
            else: # check for correct format and if present, take it
                if self.DATE_PATTERN.search(elem.text):
                    set_date(elem.text.strip())
        elif istag(elem, 'change') and elem.get('when'):
            set_date(elem.get('when').strip())

    def __format_date(self, date):
        """Bring date into the following format: YYYY-MM-dd."""
        if re.search(r"\d+-\d+-\d+", date):
            return date
        try:
            dateobj = datetime.datetime.strptime(date, "%d %B %Y")
            date = '%d-%d-%d' % (dateobj.year, dateobj.month, dateobj.day)
        except ValueError:
            pass

    def parse_dicts(self):
        raise TypeError("This class is not meant to be used directly.""")



class LocalMetaDataParser(MetaDataParser):
    """Parse meta data from XML dictionaries from a local fs path."""
    def __init__(self, name, xml):
        super().__init__(name, xml)

def get_meta_from_xml(path):
    """Parse meta data for all dictionaries in the FreeDict Root. Returns a list
    of Dictionary() objects."""
    dictionaries = []
    dict_pattern = re.compile(r'^[a-z]{3}-[a-z]{3}$')
    for item in os.listdir(path):
        full_path = os.path.join(path, item)
        matched = dict_pattern.search(item)
        if not matched: # eng-hun and hun-eng end on .header; auto-generated
            continue # is not a dictionary
        # append dictname.tei:
        full_path = os.path.join(full_path, item) + '.tei'
        if not os.path.exists(full_path):
            full_path += '.header'
        if not os.path.exists(full_path):
            raise FileNotFoundError("For dictionary %s no dictionary file was found, assumed path: %s" \
                            % (item, os.path.abspath(full_path)))
        with open(full_path, 'rb') as f:
            dparser = LocalMetaDataParser(item, f)
            dparser.parse()
            dictionaries.append(dparser.dictionary)
    return dictionaries





"""Tokenizer for text-only dictionary formats.

To use this library, add your dictionary importer as a subdirectory to
importers/. Assuming you have an immporter in importers/foo, you would add the
following snippet to your foo.py to use this module:

    from os.path import abspath, dirname
    import sys

    sys.path.append(dirname(dirname(abspath(sys.argv[0]))))

    import tokenizer

For the usage, please see the functions and classes defined in this module."""

import enum
import inspect


class ChunkType(enum.Enum):
    """The tokenizer parses individual chunks and emits tuples with type
    information and chunk content. The first entry is the chunk type, as defined
    by this enum."""
    Word = 0
    Paren = 1 # parenthesized expressions ()
    Bracket = 2 # brackets []
    Brace = 3 # "embraced" expressions: {}
    Semicolon = 4 # delimiter, e.g. homonyms
    Comma = 5 # word/definition boundary
    VerticalBar = 6 # has to be surrounded by spaces, is a delimiter as e.g. comma
    Slash = 7 # optional; example usage include abbreviations "/ETC/"

    def __repr__(self):
        return '<%s.%s>' % (self.__class__.__name__, self.name)

def space_before_slash(string, i):
    """The tokenizer can recognize chunks enclosed by slashes. To avoid to many
    false positives, only expressions with no space are supported, ATM. This
    functions checks whether a space is found before the next slash."""
    for ch in string[i+1:]: # skip first character, is the very slash
        if ch.isspace():
            return True
        elif ch == '/':
            return False
    return True # end of string reached, is like a space


FRAMING = {'(': ChunkType.Paren, '[': ChunkType.Bracket, '{': ChunkType.Brace}
CLOSING = {')':'(', '}':'{', ']':'['}
def tokenize(source, parse_slash=False):
    #pylint: disable=redefined-variable-type,too-many-branches
    """Tokenize a line (of translations) into chunks, where each chunk is a tuple of (ChunkType,
content). This only works with either the headword or the translation part,
so instead of:

    abc - a, b; c
It either parses "abc" or "a,  b; c".

Chunk types with no value (as for instance Comma) return a tuple with None as
second element.

Parenthesis have precedence over commas and ;. So

    a, b (c,;d)

will be tokenized as

    [(Word, 'a'), (Comma, None), (Word, 'b'), (Paren, 'c,;d')]

The prefix ChunkType has been omitted for readibility reasons.
This precedence rule also means that nested structures are *not* recognized,
the outer wins. This won't work:

    ([(blah)])

will result in

    [(Paren, '[(blah'), (Word, '])')]
"""
    chunks = []
    tmp_storage = [] # for unfinished chunks
    def save_parsed_chunk(ctype):
        """Save tmp_storage in chunks with the given type. This function is a short-hand."""
        nonlocal chunks, tmp_storage
        if not all(c.isspace() for c in tmp_storage):
            chunks.append((ctype, ''.join(tmp_storage).strip()))
        tmp_storage = []
    state = ChunkType.Word # current state of parser (ChunkType)
    prevchar = ''

    is_new_word = lambda: prevchar == '' or prevchar.isspace()
    for idx, ch in enumerate(source):
        if state in FRAMING.values() or (parse_slash and state == ChunkType.Slash):
            if ch in CLOSING and state == FRAMING[CLOSING[ch]]:
                save_parsed_chunk(FRAMING[CLOSING[ch]])
                state = ChunkType.Word
            elif ch == '/' and state == ChunkType.Slash and parse_slash:
                save_parsed_chunk(ChunkType.Slash)
                state = ChunkType.Word
            else: # within the brace/bracket/parethesis
                tmp_storage.append(ch)

        else: # ChunkType.Word
            # separate a enclosed expression if preceded by whitespace /
            # beginning of string; handle slash separately
            if is_new_word() and (ch in FRAMING or (parse_slash and \
                    ch == '/' and not space_before_slash(source, idx))):
                # above tests that / is followed by characters only, no spaces
                # and hence no straying slashes are detected
                if ch == '/' and parse_slash:
                    state = ChunkType.Slash
                else:
                    state = FRAMING[ch]
                if tmp_storage: # add word before, if any
                    save_parsed_chunk(ChunkType.Word)
            # only recognize /foo/ if parse_slash and if space in front
            elif ch == ',': # comma outside of parens, new word
                if tmp_storage: # not empty
                    save_parsed_chunk(ChunkType.Word)
                chunks.append((ChunkType.Comma, None))
            elif ch == ';':
                if tmp_storage: # not empty
                    save_parsed_chunk(ChunkType.Word)
                chunks.append((ChunkType.Semicolon, None))
            elif ch == '|' and (prevchar == '' or prevchar.isspace()): # space around?
                if tmp_storage: # not empty
                    save_parsed_chunk(ChunkType.Word)
                chunks.append((ChunkType.VerticalBar, None))

            else:
                tmp_storage.append(ch)
        prevchar = ch
    if tmp_storage:
        save_parsed_chunk(ChunkType.Word)
    return chunks



def split_list(mylist, delim):
    """Split a list into chunks with the given delimiter. If a delimiter is a
    function, it will get the list item as function and has to return true if a
    split should occur. Returned is an iterator, yielding a chunk at a time.
    Examples:

    >>> list(split_list([1,2,3,2,4,5,2,2,10], 2))
    [[1],[3],[4,5],[],10]]
    >>> list(split_list([]))
    []
    >>> list(split_list([1,2,3,4], 20))
    [[1,2,3,4]]"""
    if not mylist:
        return []
    gathered = []
    is_splitpoint = (delim if inspect.isfunction(delim)
                else lambda e: e == delim)
    for elem in mylist:
        if is_splitpoint(elem):
            yield gathered
            gathered = []
        else:
            gathered.append(elem)
    yield gathered




"""Everything concerned with XML processing and writing."""

from xml.etree import ElementTree as ET

class TeiHeadParser:
    """This parser uses a SAX parser to parse the header information of a TEI
    file, wrapping it in an easy-to-use interface only emitting completely
    parsed tags and aborting when a <body> tag is encountered.
    Whenever a tag is encountered, self.handle_tag(...) is called."""
    def __init__(self, input_file_object):
        self.__input = input_file_object
        # make name space available for derived classes
        self._namespace = None

    def parse(self):
        # get an iterable from XML
        context = iter(ET.iterparse(self.__input, events=("start", "end")))
        # get the root element
        _event, root = next(context)
        # extract namespace
        end = root.tag.find('}')
        if end > 0:
            self._namespace = root.tag[:end+1]

        for event, elem in context:
            if event == 'start':
                if elem.tag.endswith('body'):
                    break # do not parse body
                else: continue # skip node, not fully populated

            self.handle_tag(elem)

    def handle_tag(self, node):
        pass



def istag(elem, name):
    """Tag comparison ignoring xml namespaces."""
    return elem.tag.endswith(name)

def create_node(tag, attrs=None):
    """Create ET.Element node with specified tag and attributes."""
    e = ET.Element(tag)
    if attrs:
        e.attrib = attrs.copy()
    return e


def create_child(parent, tag, attrs):
    """Create child node and attach to parent."""
    c = create_node(tag, attrs)
    parent.append(c)


def dictionary2xml(dictionary):
    """Return the ElementNode (ElementTree) representation of this dictionary.
    Raise ValueError if a mandatory field is missing"""
    if not dictionary.is_complete():
        raise ValueError("Not all mandatory keys set.")
    downloads = dictionary.get_downloads()
    attributes = {k: v  for k, v in dictionary.get_attributes().items()
            if not v is None}
    attributes['name'] = dictionary.get_name()
    # set maintainer if none present
    if 'maintainerName' not in dictionary or not dictionary['maintainerName']:
        attributes['maintainerName'] = 'FreeDict - no maintainer assigned'
    dictionary = create_node('dictionary', attributes)
    for download in downloads:
        attrib = {'platform': str(download.format),
                'size': str(download.size),
                'date': download.last_modification_date,
                'URL': str(download), 'version': str(download.version),
                'checksum': download.hash
            }
        create_child(dictionary, 'release', attrib)
    return dictionary

def indent(elem, level=0, more_sibs=False):
    i = "\n"
    if level:
        i += (level-1) * '  '
    num_kids = len(elem)
    if num_kids:
        if not elem.text or not elem.text.strip():
            elem.text = i + "  "
            if level:
                elem.text += '  '
        count = 0
        for kid in elem:
            indent(kid, level+1, count < num_kids - 1)
            count += 1
        if not elem.tail or not elem.tail.strip():
            elem.tail = i
            if more_sibs:
                elem.tail += '  '
    else:
        if level and (not elem.tail or not elem.tail.strip()):
            elem.tail = i
            if more_sibs:
                elem.tail += '  '

def write_freedict_database(path, dicts, tools):
    """Write a freedict database to ''path``."""
    root = ET.Element('FreeDictDatabase')
    root.extend([dictionary2xml(n) for n in dicts])
    sw_node = create_node('software')
    tools_node = create_node('tools', tools)
    sw_node.append(tools_node)
    root.append(sw_node)
    indent(root)
    tree = ET.ElementTree()
    #pylint: disable=protected-access
    tree._setroot(root)
    tree.write(path, encoding="utf-8", xml_declaration=True)




"""This module contains functions to write all accompanying files for the
dictionary, e.g. license or README."""

import os
import re
import shutil
import sys
import textwrap
import xml.dom.minidom

class OutputError(Exception):
    pass

def copy_readme(input_file, output_directory):
    """When a README is found in the same directory as `input_file`, the file is
copied to OUTPUT_DIRECTORY. The input file has to be called "README.xxx-yyy" or
"README.xxx-yyy.EXT", where ext is anarbitrari file extension and xxx-yyy a ISO
6639-3 three-letter language code."""

    full_input = os.path.abspath(input_file)
    input_file = os.path.basename(input_file)
    input_dir = os.path.dirname(full_input)
    iso_code = re.search('^([a-z]{3}-[a-z]{3})', input_file)
    if not iso_code:
        raise OutputError(('%s does not follow the "xxx-yyy.EXT" '
            'naming convention.') % input_file)
    iso_code = iso_code.groups()[0]
    readme = None
    for match in (re.search('^README.([a-z]{3}-[a-z]{3}).*', f)
            for f in os.listdir(input_dir)):
        if not match:
            continue
        if match.groups()[0] == iso_code:
            readme = match.string
            break
    if readme:
        shutil.copy(os.path.join(input_dir, readme),
                os.path.join(output_directory, 'README.md'))
    else:
        raise OutputError("No README found.")


def mk_makefile(output_path, additional_files=None):
    readme = [f for f in os.listdir(output_path) if f.lower().startswith('readme')]
    if readme:
        if not additional_files:
            additional_files = []
        additional_files.append(readme[0])

    additional_files = ('' if not additional_files
        else ' '.join(additional_files))
    with open(os.path.join(output_path, 'Makefile'), 'w', encoding='utf-8') as f:
        f.write("""# The line below is really just a fallback and only works if you have got a copy
# of the tools directory at this location. It's better to set the environment
# variable in your shell.
FREEDICT_TOOLS ?= ../../tools
DISTFILES = COPYING %s \
        freedict-P5.xml freedict-P5.rng freedict-P5.dtd freedict-dictionary.css\
        INSTALL Makefile NEWS
# do not generate phonemes
supported_phonetics =

include $(FREEDICT_TOOLS)/mk/dicts.mk
""" % additional_files)

def reindent_xml(file_name):
    """Reindent an XML file with Python's minidom package. Please note that this
    will likely use a large amount of main memory."""
    try:
        tree = xml.dom.minidom.parse(file_name)
        with open(file_name, 'w', encoding='utf-8') as f:
            reindented = tree.toprettyxml(indent="  ")
            # avoid multiple empty lines, inserted by minidom
            f.write('\n'.join(l for l in reindented.split('\n') if l.strip()))
    except MemoryError:
        print("Unable to reindent file using Python's minidom module.")
        if shutil.which('xmllint'):
            print("Using xmllint for reindentation.")
            ret = os.system('xmllint --format {0} > {0}.tmp'.format(file_name))
            if ret:
                print("Aborting.")
                sys.exit(21)
            with open(file_name + '.tmp', encoding='utf-8') as infile:
                with open(file_name, 'w', encoding='utf-8') as out:
                    out.write(infile.read())
            os.remove(file_name + '.tmp')
        else:
            print('\n'.join(textwrap.wrap(("Please install xmllint as a fallback (this "
                "script will automatically use it) or use some other magic to "
                "increase the available main memory."), 78)))
            sys.exit(9)





"""This file imports the epo-eng dictionary from
#!/usr/bin/env python3
<http://www.denisowski.org/Esperanto/ESPDIC/espdic_readme.htm>. Please see the
README for more information."""

import enum
import os
import sys
import urllib.request
import xml.sax.saxutils as saxutils

# allow tokenizer import
from fd_import.tokenizer import ChunkType, tokenize
import fd_import


class WordType(enum.Enum):
    Full = '' # normal words
    Prefix = 'pref' # word prefixes
    Suffix = 'suff' # :)
    Part = 'part' # "-x-", so an infix

class Definition:
    #pylint: disable=too-few-public-methods
    """Only relevant for headwords with no direct translations, they define the
    headword."""
    def __init__(self, definition):
        self.definition = definition

    def as_xml(self):
        return '<cit>\n<def>%s</def>\n</cit>' % saxutils.escape(self.definition)

class Word:
    #pylint: disable=too-few-public-methods
    """This is either a head word or a translation."""
    def __init__(self, word, usage_info=None):
        self.word = word
        # gramGrp is a list of ready-to-use xml strings, basically *one* node
        self.gramGrp, self.usage_info = self.__get_gram_info(usage_info)
        self.word_type = type

    def __repr__(self):
        return ('%s (%s)' % (self.word, self.usage_info) if self.usage_info else
                self.word)

    def __get_gram_info(self, usage_info):
        """Figure out, whether text is a collocating word or a usage hint.
        Return tuple ready-to-use XML. """
        # special cases: catch collocations
        collocations = ['for', 'of', 'up', 'to', 'in', 'on', 'by', 'upon',
            'ke', 'de', "do"] # esperanto
        if not usage_info:
            return (None, None)
        elif usage_info in collocations:
            return (['<colloc>%s</colloc>' % saxutils.escape(usage_info)], None)
        else:
            return (None, '<usg type="hint">%s</usg>' %
                    saxutils.escape(usage_info))

    def as_xml(self):
        xml = ['<cit type="trans">\n<quote>', saxutils.escape(self.word),
                '</quote>']
        if self.usage_info:
            xml.append('\n' + self.usage_info) # already escaped
        if self.gramGrp:
            xml.append('\n<gramGrp>\n%s\n</gramGrp>' % '\n'.join(self.gramGrp))
        return ''.join(xml + ['\n</cit>'])

#pylint: disable=redefined-variable-type
class HeadWord(Word):
    #pylint: disable=too-few-public-methods
    def __init__(self, word, usg=None):
        if word.startswith('-') and word.endswith('-'):
            self.type = WordType.Part
        elif word.startswith('-'):
            self.type = WordType.Suffix
        elif word.endswith('-'):
            self.type = WordType.Prefix
        else:
            self.type = WordType.Full
        self.word = word.lstrip('-').rstrip('-')
        super().__init__(self.word, usg)

    def as_xml(self):
        extent = ''
        if not self.type == WordType.Full:
            extent = ' extent="%s"' % self.type.value
        orth = '<orth%s>%s</orth>' % (extent, saxutils.escape(self.word))
        if self.gramGrp:
            orth += '\n<gramGrp>\n%s</gramGrp>' % '\n'.join(self.gramGrp)
        if self.usage_info:
            orth += '\n' + self.usage_info
        return '<form>\n%s\n</form>' % orth

def structure_translations(unordered_list):
    """Take a list of chunks (see docs of parse_meanings) and return a list of
    list of words. The outer lists holds different meanings (homonyms) and has
    in most cases only one entry. The inner list contains objects of type
    Word."""
    translations = [[]] # initialize list with one sense
    while unordered_list:
        chunk = unordered_list.pop(0)
        # test for type of chunk
        if chunk[0] == ChunkType.Paren:
            # is this an additional info for a word:
            if unordered_list and unordered_list[0][0] == ChunkType.Word:
                translations[-1].append(Word(unordered_list.pop(0)[1], chunk[1]))
            else: # no next chunk or no word
                # this is either a parenthesized expression with no word before
                # or after, then it's a definition; otherwise it's an unhandled
                # case and a bug
                if (len(translations) == 1 and not translations[0]) and \
                        (not unordered_list or unordered_list[0][0] == ChunkType.Comma): # definition
                    translations[-1].append(Definition(chunk[1]))
                else:
                    print(',',repr(translations))
                    raise ValueError(("Couldn't parse translations; tokens "
                        "parsed: %s\nCurrent token: %s\nTokens left: %s") % \
                                (translations,chunk,unordered_list))
        elif chunk[0] == ChunkType.Word:
            # word with usage info / colloc
            if unordered_list and unordered_list[0][0] == ChunkType.Paren:
                # in some cases definitions are written as "to (be in) a
                # position", this is poor markup. ignore those parenthesis and
                # add it as a whole word:
                if len(unordered_list) >= 2 and unordered_list[1][0] == ChunkType.Word:
                    translations[-1].append(Word(chunk[1] +" " + \
                            unordered_list.pop(0)[1] + " " + \
                            unordered_list.pop(0)[1]))
                else:
                    translations[-1].append(Word(chunk[1],
                        unordered_list.pop(0)[1]))
            else:
                translations[-1].append(Word(chunk[1]))
        elif chunk[0] == ChunkType.Comma: # new word starts, ignore
            pass
        elif chunk[0] == ChunkType.Semicolon: # homonym, new sense
            translations.append([])
        # add brackets and braces verbatim
        elif chunk[0] in (ChunkType.Bracket, ChunkType.Brace):
            chars = ('[]' if chunk[0] == ChunkType.Bracket else '{}')
            # readd translation in bracket/brace, verbatim
            if translations[-1]: # at least one word found
                translations[-1][-1].word += ' ' + chars[0] + chunk[1] + chars[1]
            else: # new word
                translations[-1].append(Word(chars[0] + chunk[1] + chars[1]))
        else:
            raise NotImplementedError("Unhandled case: " + repr(chunk))
    return translations

def guess_grammar_details(translations):
    """This function inspects the English translations to infer grammatical
    information. Definitions starting with "to " are verbs in the infinitive
    form, etc. The translations might get altered, for instance, to strip the
    "to " from the actual definition."""
    gram = ''
    firstword = translations[0][0]
    if isinstance(firstword, Definition):
        return (None, translations) # don't fiddle around with def's
    if firstword.word.startswith('to '):
        gram = '<pos>v</pos>'
        for deflist in translations:
            for word in deflist:
                if word.word.startswith('to '):
                    word.word = word.word[3:]
    # remove wrongly parsed colloc's
    collocs = ['<colloc>%s</colloc>' % c for c in ['to', 'a', 'an']]
    for deflist in translations:
        for word in deflist:
            if word.gramGrp: # it need to have a colloc, otherwise no checks
                for colloc in collocs:
                    if colloc in word.gramGrp:
                        idx = collocs.index(colloc)
                        del word.gramGrp[idx]
    return (gram, translations)


def translations_to_xml(translations):
    """Translate nested definition / translation (word) structure into a
    TEI:sense structure."""
    xml = []
    for definition_list in translations:
        if not definition_list:
            raise ValueError("Found empty translation list; all translations: " \
                    + repr(translations))
        xml.append('<sense>')
        xml.extend(x.as_xml() for x in definition_list)
        xml.append('</sense>')
    return xml

def write_output(input_file, base_dir, tei_skeleton, body_xml):
    """This writes the imported dictionary into the specified directory."""
    print("Writing TEI dictionary…")
    with open(tei_skeleton, 'r', encoding='utf-8') as f:
        header = f.read()
    body_start = header.find('<body>') + 6
    tei_file = os.path.join(base_dir, 'epo-eng.tei')
    if not os.path.exists(base_dir):
        os.mkdir(base_dir)
    with open(tei_file, 'w', encoding='utf-8') as f:
        f.write(header[:body_start] + '\n')
        f.write(body_xml)
        f.write(header[body_start+1:].lstrip().rstrip() + '\n')

    fd_import.output.reindent_xml(tei_file)
    if not os.path.exists(os.path.join(base_dir, 'COPYING')):
        # retrieve copyright information
        print("Downloading CC unported 3.0 license")
        with open(os.path.join(base_dir, 'COPYING'), 'wb') as f:
            req = req = urllib.request.Request(
                    'https://creativecommons.org/licenses/by/3.0/legalcode.txt',
                data=None,
                headers = {
                    'User-Agent': 'Mozilla/5.0 (X11; Linux x86_64; rv:54.0) Gecko/20100101 Firefox/54.0'
                })
            with urllib.request.urlopen(req) as u:
                f.write(u.read())

    # write Makefile
    fd_import.output.mk_makefile(base_dir, ['epo-eng.tei', 'epo-eng.patch'])


def main(input_file, tei_skeleton, output_directory):
    print("Parsing dictionary…")
    with open(input_file, 'r', encoding='utf-8') as f:
        # gnerator with word pairs; ignore indented lines (only file header,
        # ATM)
        words = (wp for wp in f.read().split('\n')
            if wp.strip() and not wp.startswith(" "))

    word_list = []
    for lnum, word_pair in enumerate(words):
        if word_pair.lstrip().startswith("#"):
            continue
        try:
            head, trans = word_pair.split(' : ')
            translations = tokenize(trans)
            translations = structure_translations(translations)
            head = tokenize(head)
            if len(head) == 2: # headword with definition
                head = HeadWord(head[0][1], head[1][1])
            else:
                head = HeadWord(head[0][1])
            word_list.append((head, translations))
        except ValueError as v:
            raise ValueError(f"line {lnum+1}: {''.join(v.args)}")

    # enrich words with grammatical info, strip "to" and other words *and*
    # convert to XML
    xml = []
    for head, trans in word_list:
        # translations might get altered, e.g. "to" is stripped
        gram, trans = guess_grammar_details(trans)
        xml += ['<entry>', head.as_xml()]
        if gram:
            xml.append('<gramGrp>\n%s\n</gramGrp>' % gram)
        xml += translations_to_xml(trans) + ['</entry>']

    write_output(input_file, output_directory, tei_skeleton, '\n'.join(xml))
    print("Done. Now it's time to copy DTD, CSS and RNG and validate the dictionary.")

def check_args():
    """command line parameter validation"""
    if len(sys.argv) != 4:
        print("Error: invalid command line parameters.")
        import textwrap
        print("Usage: %s <INPUT_FILE> <TEI SKELETON> <OUTPUT_DIRECTORY>\n    "\
                % sys.argv[0], end="")
        print('\n    '.join(textwrap.wrap(("TEI SKELETON has to be a FreeDict TEI file, with an empty "
                "body tag. The opening and closing body tags have to be on "
                "separate lines.\n"
                "The output directory may exist and files will be overwritten, "
                "as necessary."), 78)))
        sys.exit(1)

    if not os.environ['FREEDICT_TOOLS']:
        print("The environment variable FREEDICT_TOOLS is not set.")
        print("This is mandatory, so that the necessary scripts can be located.")
        print("Have a look at the FreeDict wiki for more information:")
        print("https://github.com/freedict/fd-dictionaries/wiki")
        sys.exit(4)
    for file in sys.argv[1:3]:
        if not os.path.exists(file):
            print('Sorry, but the file "%s" does not exist.' % file)
            sys.exit(2)
    dictdir = os.path.basename(os.path.abspath(sys.argv[3]))
    if not len(dictdir) == 7 or '-' not in dictdir:
        print("The output directory has to be named with the usual FreeDict naming scheme. Otherwise, the Make rules will fail.")
        sys.exit(5)

if __name__ == '__main__':
    check_args()
    main(sys.argv[1], sys.argv[2], sys.argv[3])




"""This module holds datastructures to represent a dictionary and links to their
downloads and a few function to retrieve useful information about these
objects."""
import datetime
import distutils.version
import enum
import os
import re
import urllib.request

import semver

from . import config


class Dictionary:
    """Dictionary class holding various properties of a dictionary.
    This class is intended to hold all properties of a Dictionary node in the
    resulting XML API file.

    A dictionary object consists of optional and mandatory information:

    mandatory: headwords, edition, date
    optional = maintainerName, maintainerEmail, status, sourceURL

    The date should be the date of the last release.
    Additionally, a dictionary keeps a list of downloads which may be empty.
    Each download has to be a Link() object.
    """

    def __init__(self, name):
        self.__name = name
        # mandatory dictionary information
        mandatory = ["headwords", "edition", "date"]
        self.__mandatory = {f: None for f in mandatory}
        # optional dictionary info
        self.__optional = {
            f: None
            for f in ["maintainerName", "maintainerEmail", "status", "sourceURL"]
        }
        self.__downloads = []

    def get_name(self):
        """Return name of dictionary. This is a three-letter-code followed by a
        hyphen and followed by a three-letter-code.
        Example: deu-fra"""
        return self.__name

    def add_download(self, link):
        """Add a link (of type Link) to the linst of downloadss."""
        if not isinstance(link, Link):
            raise TypeError("Link must be of type Link()")
        self.__downloads.append(link)

    def get_downloads(self):
        """Return all download links."""
        return self.__downloads

    def __getitem__(self, key):
        """Transparently select a key from either optional or mandatory keys."""
        if key in self.__mandatory:
            return self.__mandatory[key]
        if key in self.__optional:
            return self.__optional[key]
        raise KeyError(key)

    def __contains__(self, key):
        try:
            self.__getitem__(key)
        except KeyError:
            return False
        else:
            return True

    def __setitem__(self, key, value):
        if key in self.__mandatory:
            self.__mandatory[key] = value
        elif key in self.__optional:
            self.__optional[key] = value
        else:
            raise KeyError(key)

    def get_mandatory_keys(self):
        """Return all mandatory keys."""
        return self.__mandatory.keys()

    def is_complete(self):
        """Return true if all mandatory fields are set, else false."""
        return not self._get_missing_keys()

    def _get_missing_keys(self):
        """Return list of keys which haven't been set yet but which are
        mandatory. Empty list means everything has been set."""
        return [k for k in self.__mandatory if self.__mandatory[k] is None]

    def update(self, other):
        """This method works like the .update method on a dictionary, but it
        raises an exception whenever an unknown key is found in the supplied
        dictionary."""
        if not hasattr(other, "__getitem__") or not hasattr(other, "keys"):
            raise TypeError("Object must provide methods keys() and __getitem__.")
        for key in other.keys():
            self[key] = other[key]

    def get_attributes(self):
        """Return all attributes which make up the dictionary node in the
        FreeDict XML. If mandatory attributes are not set, this function WILL
        NOT raise an exception, consequently, is_complete() must be called
        beforehand. Unset values are None.
        Hint: the name is not contained, use get_name() instead."""
        attributes = self.__mandatory.copy()
        attributes.update(self.__optional)
        return attributes


# only match x.x, x.x.x, x-y-z, x-z
DICTIONARY = "([a-z]{3}-[a-z]{3})"


class DownloadFormat(enum.Enum):
    """The download format, consisting both of the platform and the archive
    format. Some formats might not have a archive format, though."""

    DictTxz = re.compile(r"freedict-%s-(.*?).dictd.tar.xz" % DICTIONARY)
    Slob = re.compile(r"freedict-%s-(.*?).slob" % DICTIONARY)
    StardictTxz = re.compile(r"freedict-%s-(.*?).stardict.tar.xz" % DICTIONARY)
    Source = re.compile(
        r"freedict-%s-(.*?).src.(?:zip|tar.bz2|tar.gz|tar.xz)" % DICTIONARY
    )

    # legacy formats
    DictTgz = re.compile(r"freedict-%s-(.*?).dictd.tar.gz" % DICTIONARY)
    DictBz2 = re.compile(r"freedict-%s-(.*?).dictd.tar.bz2" % DICTIONARY)
    Dic = re.compile(r"freedict-%s-(.*?).dic.tar.*" % DICTIONARY)

    @staticmethod
    def get_type(file_name):
        """This function allows to get the correct enum value from a given
        file_name. The file name is parsed and the corresponding enum value
        returned. If the format could not be extracted, None is returned."""
        if file_name.endswith(".sha512"):
            return None
        format = DownloadFormat.Source
        for format in DownloadFormat:
            if format.value.search(file_name):
                return format
        return None

    def __str__(self):
        """Return a string representation of the enum value, as used in the type
        attribute of the release tag in the FreeDict XML API."""
        formats = {
                DownloadFormat.Source: "src",
                DownloadFormat.DictTgz: "dict-tgz",
                DownloadFormat.DictBz2: "dict-bz2",
                DownloadFormat.DictTxz: "dictd",
                DownloadFormat.Slob: "slob",
                DownloadFormat.StardictTxz: 'stardict'
            }
        try:
            return formats[self]
        except KeyError:
            raise ValueError("Unsupported format: " + self.name)


class Link:
    """Represent a (download) link.

    The link is made of of multiple parts. The hostname and base URI is taken
    from variables defined in the module config.
    The given path is assumed to exist on disk, so that the file can be
    determined.
    Additionally to the given link path, a link also saves the information about
    the file format (dict-bz2 or source) and also the version of the
    dictionary.
    It also mandates a hash for verification of the file."""

    def __init__(self, path, format, version, sha):
        self.path = path
        self.format = format
        self.version = version
        self.size = -1
        self.last_modification_date = "NONE"  # YYYY-MM-dd
        self.hash = sha

    def __str__(self):
        """Get a download link."""
        # split the path into chunks, url-quote them
        path = tuple(map(urllib.request.quote, self.path.split(os.sep)))
        if len(path) < 3:
            raise ValueError(
                "Required is a path with the structure LongName/version/filename"
            )
        path_base = config.RELEASE_HTTP_BASE
        if not path_base.endswith("/"):
            path_base += "/"
        return "https://{}{}{}/{}/{}".format(
            config.PROJECTHOME_HOST, path_base, path[-3], path[-2], path[-1]
        )


def normalize_version(vers_string):
    """Make a semantic version out of a less strict version number , e.g. 0.5 to
    0.5.0."""
    try:
        return semver.parse(vers_string)
    except ValueError:
        return
    semver.parse(".".join(distutils.version.StrictVersion(vers_string).version))


def mklink(full_path, format, version, sha):
    """Create a Link object with all the required information, i.e. file size.
    It queries the referenced `full_path` for its size and last modification
    date. It doesn't care whether it's actually on the file system or mounted
    with e.g. sshfs."""
    chunks = full_path.split(os.sep)
    path = "{}/{}/{}".format(chunks[-3], chunks[-2], chunks[-1])
    link = Link(path, format, version, sha)
    # get file size
    link.size = os.path.getsize(full_path)
    # get last modification date
    link.last_modification_date = datetime.datetime.fromtimestamp(
        os.path.getmtime(full_path)
    ).strftime("%Y-%m-%d")
    return link



#!/usr/bin/python
# Written by Petergozz, Jan 2004
#
# micha137: renamed from dict2xml.py
#
## todo GPL 2+ ##
##### THIS IS ALPHA LEVEL NOT FOR PRODUCTION USE ###########
####### Requires Python2.3 or later ###########################
##  TODO will need TEI header > proper tei stuff too !!
##	d2X_write_tei_header() (not here)
## TODO add detect for .dz files and uncompress
## (if tools on board if not .. explore the gzip modules :)
## dz is a modded version of gzip so _might_ be  doable ?


import sys
import time
import os
import string
import re
#cool new way to do getopts :)
#import optparse
from optparse import OptionParser, OptionGroup



#
# Globals
#
VERSION = "-0.1.1"
chatty = None
app = os.path.basename(sys.argv[0])
start_time = time.asctime()
#
# regex defs (pre-compiles) these are used in d2x_format_xml
#
rex_hdwd = re.compile('^\w.*$') #Headword  starts with anything not a white space
rex_descpt = re.compile('^\s\s+.*$') #Description starts with more than one white space
## TODO add matches for parts of speech pronunciation etc. here hmm more command line options ...

## TODO add matches for file names here (to autogen out names)

## TODO add matches for 00-data etc for dictd headers (possibly)
def d2x_getInput():
	d2x_usage = '%prog -f dictfile [options]\n\n Defaults are provided for _everything_ except the dictfmt FILE to read from '
	cl_parser = OptionParser(d2x_usage, version="%prog"+VERSION )

	cl_parser.add_option("-f", "--file", type="string", action="store",  dest="readfile",  help="read dictfmt file  from FILENAME" )
	cl_parser.add_option("-v", "--verbose", action="store_true", dest="verbose",  help="Tell me whats going on.  ")
	cl_parser.add_option("-o", "--out", type="string", action="store", dest="writefile", default="dicttei.xml",  help="write TEI/XML format file to FILENAME" )
	groupDocT = OptionGroup( cl_parser, "Advanced Options for changing the DOCTYPE", "Use these to set a doctype string that works for your system")
	groupDocT.add_option("-s", "--dtdsys", type="string", action="store", dest="DTDsys",default='http://www.tei-c.org/Guidelines/DTD/tei2.dtd' , help="set System DTD  to PATH.  NB: If your not using an XML/SGML catalog system you should set this to: /your/path/to/tei2.dtd" )
	groupDocT.add_option("-p", "--dtdpub", type="string",action="store", dest="DTDpub",default='-//TEI P4//DTD Main Document Type//EN', help="set public DTD to \"Formal Public Identifier\"  NB: You _will_ need to quote it" )
	groupDocT.add_option("-t", "--dtdtype", type="string", action="store", dest="DTDtype", default="TEI.2", help="set non default DOCTYPE [TEI.2] " )
	cl_parser.add_option_group( groupDocT )
	groupXML = OptionGroup( cl_parser, "Advanced options for altering the default XML header.", "Use these if you need to change the defaults. There are no single switch options for these." )

	groupXML.add_option("--xmlver" , type="string", action="store", dest="XMLver", default='1.0', help="Set XML version attribute. [\"1.0\" ]" )
	groupXML.add_option("--xmllang", type="string", action="store", dest="XMLlang", default='en', help="set the XML code language attribute. [en]")
	groupXML.add_option("--xmlstand", type="string", action="store", dest="XMLstand", default='no', help="set the XML \"standalone\"  attribute. [no]")
	groupXML.add_option("--xmlenc", type="string", action="store", dest="XMLenc", default='utf-8', help="set the XML character ISO code attribute. [utf-8] \n ")
	cl_parser.add_option_group( groupXML )
	## TODO a really quiet option and a logging option and a dotfile prefs section and group the options so they don't scare the crap out of innocent bystanders.
	(cl_options, cl_args)  = cl_parser.parse_args()

	#pull the exports out of the "getopt"
	dictFileIN = cl_options.readfile
	teiFileOut = cl_options.writefile
	dtdType = cl_options.DTDtype
	dtdSys = cl_options.DTDsys
	dtdPub = cl_options.DTDpub
	chat = cl_options.verbose
	xml_v = cl_options.XMLver
	xml_lang = cl_options.XMLlang
	xml_stand = cl_options.XMLstand
	xml_enc = cl_options.XMLenc

	# catch-me's here
	if len(cl_args) << 1: ## this still broken i will fix later
		cl_parser.error("We need at least one thing to do.\n\n Have you supplied a file name for reading ?\n <::For help type::> "+ app +" -h")
	elif dictFileIN == None :
		print app +"      ::>   No input file  <::\n"
		cl_parser.print_help()
		sys.exit(0)
	else:
		print app +"        Reading from:::> "+ dictFileIN + "  <::\n"
		print app +"        Writing to:  ::> "+ teiFileOut +" <::\n"

	#Test for verbosity
	# (damm and blast this is clunky)

	print app+" REMINDER ::> This is Alpha level software ! <::"
	print app+ VERSION +" !!!!!!!!!!!!  not for production use !!!!!!!!!!!!!!!!"
	if chat == True :
		print "command line options   :", cl_options
		chatty = "Y"
		print "Chat mode is on" +chatty
	else :
		chatty = "N"
		print app +" Chat mode off"



	#
	#Now get to work
	#call the workhorses up
	#
	d2x_write_prolog( app, teiFileOut, dtdType, dtdPub, dtdSys, xml_v, xml_lang, xml_stand, xml_enc, chatty  )
	d2x_format_xml( dictFileIN, teiFileOut, chatty  )
	return()


def d2x_write_prolog( this_app, fout, doc_t,  doc_type_pub, doc_type_sys,  xml_v, xml_lang, xml_stand, xml_enc,  chatty ):
	if chatty == "Y":
		print "entered write prolog function"

	xmlfile = file(fout, "w+")

	if chatty == "Y":
		print "Writing to ::> ", xmlfile


	# prolog is just a concat of all the following:
	doc_type =  '<!DOCTYPE  '+ doc_t+ '  PUBLIC \"'+ doc_type_pub +'\"  \"' + doc_type_sys +'\" [ \n<!ENTITY % TEI.XML            "INCLUDE" >\n<!ENTITY % TEI.dictionaries \"INCLUDE\" > \n]>\n<!--this file  auto generated on   ' +start_time +'   by ' + this_app + VERSION +' \n     please edit and rename  --> '
	xml_head = '<?xml version=\"'+xml_v+'\"  encoding=\"'+xml_enc+'\"  lang =\"'+xml_lang+'\" standalone=\"'+xml_stand+'\" ?>'
	#
	#So putting it all together we get
	#
	prolog = xml_head+'\n'+doc_type+'\n\n'
	if chatty == "Y" :
		print(prolog)

	xmlfile.write( prolog )
	xmlfile.close() # this seems safer and dumps the buffer (we need a lot of ram for big files)


def d2x_format_xml(fin, fout, chatty ) :
	"""d2x_format_xml()

	takes a dictd format file and wraps it in TEI print dictionary xml tags.
	Command line options exist for most sgml, XML attributes and file in and out names.
	Defaults are supplied for all but the file in name.
	 """

	dictfmt = file( fin, 'r+',1 ) #open file in read and write mode line buffed only

	xmlfile = file( fout, 'a' ) # reopen the output file for appending

	# read all of dictfmt file to a list (as it only has new lines to differentiate with)
	dictlist = dictfmt.read()
	##TODO break into 100 line (+/- 8K) blocks use seek to increment through the whole file?
	# now split the buffer by "2 or more new lines"
	dictarray = dictlist.split('\n\n')

	# TODO make a spinner or % readout(after you improve cache and speed)

	for record in dictarray:
		recSub1 = re.sub('^\n', "" , record)#tidy any leading newlines
		recSub2  = re.sub('\t', '    ', recSub1) # replace tabs with 4 spaces
		recSub3 = recSub2+'\n'+'</entry>' # tag the true end
		sub_string = recSub3.split('\n')
	#
	#note do not strip leading space from defs (yet)
	#
		# it should now be the case that headwords start on "col one" and defs etc don't
		xmlfile.write('\n<entry>')
		for field in sub_string:
			if chatty == "Y":
				print "found field"

			match_H =  rex_hdwd.match( field )
			match_D =  rex_descpt.match( field )
			match_End = re.search('</entry>', field)
			if  match_H :
				if chatty == "Y":
					print 'Headword Match found: ', match_H.group()

				xmlfile.write('\n<form><orth>' )
				xmlfile.write(match_H.group())
				xmlfile.write('</orth></form>')
			elif match_D :
				if chatty == "Y":
					print 'Description Match found: ',  match_D.group()

				xmlfile.write( '\n<def>')
				xmlfile.write( match_D.group() )
				xmlfile.write ('</def>')
			elif match_End :
				if chatty == "Y":
					print 'end entry'

				xmlfile.write ('\n</entry>')
			else:
				if chatty == "Y":
					print 'No match'

#
#detect mode of operation and gather an environment etc
#we actually start from here if called to execute
#
if  __name__ == "__main__":
	d2x_getInput() # NTS this is not C
	print app+ ": End Run"




#!/usr/bin/env python3
"""
This script attempts to deduplicate entries in a dictionary or to detect whether
duplicates are present.
In the normal operation mode, the questionable entries are removed, the result
is written to a file-dedup.py and then it is converted to the C5 format. From
there, a diff with the original format is done. With the diff, the human in
front of this very computer can then reason about the changes.

There is also a mode where the script tries to find a duplicated translation and
exits upon the first match.

Dependencies: xsltproc, diff python >= 3.4.

The script tries to preserve the original XML, but may do the following changes:

-   Manual sense enumeration (using the attribute `n`) may be stripped.
    Because of the removal of doubled senses, the enumeration might be broken
    and it's automatically inserted by the output formatters, anyway.
-   XML declarations may get lost, feel free to add support for it. Comments
    work fine.
"""

import argparse
import io
import itertools
import os
import shlex
import shutil
import sys
import xml.etree.ElementTree as ET

# TEI name space, Python's parser doesn't handle them nicely
TEI_NS = '{http://www.tei-c.org/ns/1.0}'
NS = {'t': TEI_NS.strip('{}')}
# findall/iter with TEI namespace removed
findall = lambda x,y: x.findall(TEI_NS + y)
tei_iter = lambda x,y: x.iter(TEI_NS + y)

class HelpfulParser(argparse.ArgumentParser):
    """Unlike the super class, this arg parse instance will print the error it
    encountered as well as the complete usage of the program. It will also
    redirect the usage to a output formatter."""
    def __init__(self, name, description=None):
        super().__init__(prog=name, description=description)

    def error(self, message):
        """Print error message and usage information."""
        sys.stderr.write('Error: ' + message + '\n')
        self.print_help()
        sys.exit(2)


class CommentedTreeBuilder(ET.TreeBuilder):
    """A TreeBuilder subclass that retains XML comments from the source. It can
    be treated as a black box saving the contents before and after the root tag,
    so that it can be re-added when writing back a XML ElementTree to disk. This
    is necessary because of ElementTree's inability to handle declarations
    nicely."""
    def comment(self, data):
        self.start(ET.Comment, {})
        self.data(data)
        self.end(ET.Comment)

# These helpers allow the identification of similar nodes (comparison of quote
# or usg)
def nodes_eq(node1, node2, tag=None):
    """Check whether both given nodes contain equal text. If tag is given, not
    the nodes themselves, but the given child nodes are compared."""
    if not tag:
        tag = node1.tag
    node1 = node1.find('.//' + TEI_NS + tag)
    node2 = node2.find('.//' + TEI_NS + tag)
    if node1 is None and node2 is None:
        return True # means equality, differences are searched
    if node1 is None or node2 is None:
        return False
    return node1.text == node2.text

usages_match = lambda n1, n2: nodes_eq(n1, n2, 'usg')
def translations_of(sense):
    trans = sense.findall('.//t:quote', NS)
    if not trans:
        trans = sense.findall('.//t:def', NS)
    return trans

def rm_doubled_senses(entry):
    """Some entries have multiple senses. A few of them are exactly the same,
    remove these.
    This function returns True if an element has been altered"""
    senses = list(findall(entry, 'sense'))
    if len(senses) == 1:
        return

    # obtain a mapping from XML node -> list of words within `<quote>…</quote>`
    senses = {sense: translations_of(sense) for sense in senses}
    changed = False
    # pair each sense with another and compare their content
    for s1, s2 in itertools.combinations(senses.items(), 2):
        # both sets match
        if not len(s1[1]) == len(s2[1]):
            continue
        if not set(t.text for t in s1[1]) == set(t.text for t in s2[1]):
            continue # don't match
        # two senses are *exacctly* identical?
        try:
            entry.remove(s2[0]) # remove sense from entry
            changed = True
        except ValueError: # already removed?
            pass
    return changed

def node_is_empty(node):
    """Check that this node and all its children contain no text (apart from
    white space)."""
    if node.text.strip():
        return False
    for child in node.getchildren():
        if not node_is_empty(child):
            return False
    return True

def rm_empty_nodes(entry):
    """This function removes nodes which have no text and no children and are
    hence without semantic. It also resets the fixed counters on sense, since
    they are generated by the output formatter anyway.
    This function returns True if an empty node has been removed."""
    changed = False
    # sometimes parent nodes are empty after their empty children have been
    # removed, so do this three times (won't work with deeper nestings…)
    for _ in range(0, 2):
        changed_here = False
        nodes = [(None, entry)]
        for parent, node in nodes:
            if node_is_empty(node):
                if parent:
                    parent.remove(node)
                    changed_here = changed = True
            else:
                nodes.extend((node, c) for c in node.getchildren())
        if not changed_here:
            break # nothing removed, no further iterations
        else:
            changed = True
    # try to strip enumeration of senses they aren't adjacent anymore; map
    # sense: numeration first or discard if no numbering
    sense_numbers = {sense: int(sense.attrib.get('n'))
            for sense in tei_iter(entry, 'sense') if sense.attrib.get('n')}
    if sense_numbers:
        # if not all values between min and max sense number, enumeration is
        # broken and can be stripped
        if not all(n in sense_numbers.values()
                for n in range(1, max(iter(sense_numbers.values())))):
            for node in sense_numbers:
                # strip manual enumeration, handled by output formatters and might
                # be wrong after node removal
                del node.attrib['n']
                changed = True
    return changed

def rm_doubled_quotes(entry):
    """Some entries have doubled quotes (translations) within different senses.
    Remove the doubled quotes.
    This function return True, if the entry has been modified."""
    senses = list(findall(entry, 'sense'))
    # add quote elements
    senses = [(s, cit, q)  for s in senses for cit in findall(s, 'cit')
            for q in findall(cit, 'quote')]
    if len(senses) <= 1:
        return
    changed = False
    # pair each sense with another and compare their content
    for trans1, trans2 in itertools.combinations(senses, 2):
        sense1, _cit1, quote1 = trans1
        sense2, cit2, quote2 = trans2
        # translation could have been removed by a previous pairing
        if quote1.text == quote2.text and usages_match(sense1, sense2):
            try:
                cit2.remove(quote2)
            except ValueError:
                continue # already removed
            changed = True
    return changed

def exec(command):
    """Execute a command or fail straight away."""
    ret = os.system(command)
    if ret:
        sys.stderr.write("Process exited with %i: %s" % (ret, command))
        sys.exit(ret)

#pylint: disable=too-few-public-methods
class XmlParserWrapper:
    """This thin wrapper guards the parsing process.  It manually finds the TEI
    element and copies everything before and afterwards *verbatim*. This is due
    to the inability of the ElementTree parser to handle multiple "root
    elements", for instance comments before or after the root node or '<!'
    declarations.
    """
    def __init__(self, file_name):
        with open(file_name, encoding='utf-8') as file:
            content = file.read()
        tei_start = content.find('<TEI')
        if tei_start < 0:
            raise ValueError("Couldn't find string `<TEI` in the XML file.  Please extend this parser.")
        self.before_root = content[:tei_start]
        content = content[tei_start:]
        tei_end = content.find('</TEI>')
        if tei_end < 0:
            raise ValueError("Couldn't find `</TEI>` in the input file, please extend the parser.")
        tei_end += len('</TEI>')
        self.after_root = content[tei_end:]
        content = content[:tei_end]
        parser = ET.XMLParser(target = CommentedTreeBuilder())
        try:
            parser.feed(content)
        except ET.ParseError as e:
            sys.stderr.write("Error while parsing input file\n")
            sys.stderr.write(str(e) + '\n')
            sys.exit(15)

        self.root = parser.close()

    def write(self, file_name):
        """Write the XML element tree to a file, with hopefully a very similar
        formatting as before."""
        tree = ET.ElementTree(self.root)
        in_mem = io.BytesIO()
        tree.write(in_mem, encoding="UTF-8")
        in_mem.seek(0)
        with open(file_name, 'wb') as file:
            file.write(self.before_root.encode('UTF-8'))
            file.write(in_mem.read())
            file.write(self.after_root.encode('UTF-8'))
            if not self.after_root.endswith('\n'):
                file.write(b'\n')


def main():
    parser = HelpfulParser("deduplicator", description=("Fnd and remove "
        "duplicated translations and empty TEI nodes"))
    parser.add_argument("-s", "--detect_changes", dest="detect_changes",
              help=("check whether duplicates or empty nodes can be detected "
                  "and exit with exit code 42 if the first change would "
                  "need to be made"),
                  action="store_true", default=False)
    parser.add_argument('dictionary_path', help='input TEI file', nargs="+")
    args = parser.parse_args(sys.argv[1:])

    # register TEI name space without prefix to dump the *same* XML file
    ET.register_namespace('', 'http://www.tei-c.org/ns/1.0')
    dictionary_path = args.dictionary_path[0]
    tree = XmlParserWrapper(dictionary_path)
    changed = False
    if args.detect_changes: # abort if first change detected
        for entry in tei_iter(tree.root, 'entry'):
            changed = changed or rm_doubled_senses(entry)
            # this one is dangerous
            #changed = changed or rm_doubled_quotes(entry)
            # the processing above might leave empty parent nodes, remove those
            changed = changed or rm_empty_nodes(entry)
            if args.detect_changes and changed:
                print(("E1: Found duplicated entries or empty XML nodes. Try "
                        "`make rm_duplicates`."))
                sys.exit(42)
    else: # always apply changes
        for entry in tei_iter(tree.root, 'entry'):
            changed1 = rm_doubled_senses(entry)
            changed2 = rm_doubled_quotes(entry)
            # the processing above might leave empty parent nodes, remove those
            changed3 = rm_empty_nodes(entry)
            if changed1 or changed2 or changed3:
                changed = True
        if args.detect_changes and changed:
            print(("E1: Found duplicated entries or empty XML nodes. Try "
                    "`make rm_duplicates`."))


    if changed:
        output_fn = os.path.join('build', 'tei',
                dictionary_path.replace('.tei', '-dedup.tei'))
        exec('mkdir -p build/tei')
        tree.write(output_fn)
        # get a human-readable diff of the changes
        c5 = lambda x: shlex.quote(x.replace('.tei', '.c5'))
        shutil.copy('freedict-P5.dtd', os.path.dirname(output_fn))
        exec('xsltproc $FREEDICT_TOOLS/xsl/tei2c5.xsl %s > %s' % (output_fn,
            c5(output_fn)))
        # convert original dictionary to c5
        exec('make --no-print-directory build-dictd')
        # execute diff without checking the return type
        if not shutil.which('less'):
            exec('diff -u build/dictd/%s %s' % (c5(dictionary_path), c5(output_fn)))
        else:
            os.system('diff -u build/dictd/%s %s | less' % (c5(dictionary_path), c5(output_fn)))
        if input("Do you want to overwrite %s with the new version (y|n): " % \
                dictionary_path) == 'y':
            shutil.copy(output_fn, dictionary_path)


if __name__ == '__main__':
    main()



#!/usr/bin/env python3

import sys
import random
from xml.etree import ElementTree as etree

from ankisync.apkg import Apkg

NS = 'http://www.tei-c.org/ns/1.0'
NS_MAP = {'': NS, 't': NS}


def ns(name):
    return '{%s}%s' % (NS, name)


TAG_ENTRY = ns('entry')
TAG_HEADER = ns('teiHeader')


def parse_entry(element):
    orths = element.findall('./t:form//t:orth', NS_MAP)
    cits = element.findall('./t:sense//t:cit/t:quote', NS_MAP)
    return orths[0].text, ', '.join(c.text for c in cits)


assert len(sys.argv) == 2, 'Usage: tei2anki.py FROM-TO.tei'
input_filename = sys.argv[1]
output_filename = input_filename.replace('.tei', '.apkg')

entries = []
with open(input_filename) as input_file:
    for _, element in etree.iterparse(input_file):
        if element.tag == TAG_HEADER:
            deck_name = element.find('./t:fileDesc/t:titleStmt/t:title', NS_MAP).text
        if element.tag == TAG_ENTRY:
            entries.append(parse_entry(element))

with Apkg(output_filename) as apkg:
    model_id = apkg.init(
        first_model=dict(
            name='default model',
            fields=['source', 'target'],
            templates={
                'Forward': ('{{source}}', '{{source}} <hr id=answer> {{target}}'),
            }
        ),
        first_deck=deck_name,
        first_note_data=False,
    )
    for source, target in random.sample(entries, 100):
        apkg.add_note({
            'modelName': 'default model',
            'deckName': deck_name,
            'fields': {
                'source': source,
                'target': target,
                }
            })
